#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 12 11:23:20 2025

@author: ncro8394
"""


from os import listdir, mkdir, path
from pandas import concat, DataFrame
from wonambi import Dataset, graphoelement
from wonambi.attr import Annotations
from wonambi.trans import fetch
import shutil
from yasa import rem_detect
from copy import deepcopy
from ..utils.logs import create_logger
from ..utils.load import load_sessions, load_stagechan, load_emg, load_eog
from ..utils.squid import infer_eog

class remora:
    
    """ Rapid Eye Movement Oscillation Recognition Algorithm (R.E.M.O.R.A)

        This module runs automated sleep staging with the option of using
        previously published staging algorithms:
            1. Vallat et al. (2020) - YASA
            2. 
        
    """   
    
    def __init__(self, rec_dir, xml_dir, out_dir, eog_chan, 
                 ref_chan = None, rater = None, grp_name = 'eeg',
                 reject_artf = ['Artefact', 'Arou', 'Arousal'],
                 subs='all', sessions='all', 
                 tracking = None):
        
        self.rec_dir = rec_dir
        self.xml_dir = xml_dir
        self.out_dir = out_dir
        self.eog_chan = eog_chan
        self.ref_chan = ref_chan
        self.rater = rater
        self.grp_name = grp_name
        self.reject = reject_artf
        self.subs = subs
        self.sessions = sessions
        
        if tracking == None:
            tracking = {}
        self.tracking = tracking
        
        
    def detect_rems(self, method, amplitude = (50, 325), duration = (0.3, 1.5), 
                          stage = ['REM'], cycle_idx = None, filetype = '.edf', 
                          logger = create_logger('Detect sleep stages')):
         
         ''' Automatically detects sleep stages by applying a published 
             prediction algorithm.
         
             Creates a new annotations file if one doesn't already exist.
         
             Parameters
             ----------
             
             method      ->   str of name of automated detection algorithm to 
                              detect staging with. Currently only 'YASA' 
                              is supported. 
                              (https://raphaelvallat.com/yasa/generated/yasa.rem_detect.html)
                              
             qual_thresh ->   Quality threshold. Any stages with a confidence of 
                              prediction lower than this threshold will be set 
                              to 'Undefined' for futher manual review.
    
         
         '''
         
         flag = 0
         tracking = self.tracking
         
         logger.info('')
         logger.debug(rf"""Commencing rapid eye movements detection... 
                      
                      
                                        
                                       /`·.¸       ¸..¸
                     ¸.···´"·........."..···'····'´    '·....¸¸ 
                    >  © ):´;                         ¸.¸¸¸    `````·.¸¸¸·´\ 
                     `·.¸ `·                   ¸.···´´     ````````````. ¸ )     
                         ```''\¸.'´´´´´´´´´´´´                         \  /     
                                                                        '  
                                 
                     Rapid Eye Movement Oscillation Recognition Algorithm 
                     (R.E.M.O.R.A)
                     
                     Using method: {method}
                     
                                                     """,)
         ### 1. First we check the directories
         # a. Check for output folder, if doesn't exist, create
         if path.exists(self.out_dir):
                 logger.debug("Output directory: " + self.out_dir + " exists")
         else:
             mkdir(self.out_dir)
         
         # b. Check input list
         subs = self.subs
         if isinstance(subs, list):
             None
         elif subs == 'all':
                 subs = listdir(self.rec_dir)
                 subs = [p for p in subs if not '.' in p]
         else:
             logger.error("'subs' must either be an array of subject ids or = 'all' ")  
             return
         
         ### 2. Begin loop through dataset
        
         # a. Begin loop through participants
         subs.sort()
         for i, sub in enumerate(subs):
             # b. Begin loop through sessions
             flag, sessions = load_sessions(sub, self.sessions, self.rec_dir, flag, 
                                      logger, verbose=2)
             for v, ses in enumerate(sessions):
                 logger.info('')
                 logger.debug(f'Commencing {sub}, {ses}')


                 ## c. Load recording
                 rdir = f'{self.rec_dir}/{sub}/{ses}/eeg/'
                 try:
                     edf_file = [x for x in listdir(rdir) if x.endswith(filetype)][0]
                     dset = Dataset(f'{rdir}/{edf_file}/')
                 except:
                     logger.warning(f'No input {filetype} file in {rdir}')
                     break
                 
                 # d. Load/create/read for annotations file
                 xdir = f'{self.xml_dir}/{sub}/{ses}'
                 xml_file = [x for x in listdir(xdir) if x.endswith('.xml')][0]
                 if not path.exists(f'{self.out_dir}/{sub}'):
                     mkdir(f'{self.out_dir}/{sub}')
                 if not path.exists(f'{self.out_dir}/{sub}/{ses}'):
                      mkdir(f'{self.out_dir}/{sub}/{ses}')
                 backup = f'{self.out_dir}/{sub}/{ses}/'
                 backup_file = (f'{backup}{sub}_{ses}.xml')
                 if not path.exists(backup_file):
                     shutil.copy(f'{xdir}/{xml_file}', backup_file)
                 else:
                     logger.warning(f'Annotations file already exists for {sub}, {ses}, any previously detected events will be overwritten.')

                 annot = Annotations(backup_file, rater_name = self.rater)
                 
                 ## e. Get sleep cycles (if any)
                 if cycle_idx is not None:
                     all_cycles = annot.get_cycles()
                     cycle = [all_cycles[y - 1] for y in cycle_idx if y <= len(all_cycles)]
                 else:
                     cycle = None
                 
                 # f. Load chans
                 pflag = deepcopy(flag)
                 flag, eog_chan = load_eog(sub, ses, self.eog_chan,
                                      flag, logger)
                 if flag - pflag > 0 or len(eog_chan) != 2:
                     logger.debug("Unable to obtain EOG from tracking sheet. Ensure there are 2 EOG channels listed under 'eog'.")
                     logger.debug("Inferring EOG directly from file instead.")
                     eog_chan = infer_eog(dset, logger)
                     if not eog_chan:
                          logger.warning(f'Unable to infer EOG. Skipping {sub}, {ses}...')
                          break
                 if not isinstance(eog_chan, list):
                      eog_chan = [eog_chan]
                      
                 # Load ref
                 flag, chanset = load_stagechan(sub, ses, eog_chan, self.ref_chan,
                                               flag, logger)
                 ref_chan = chanset[eog_chan[0]]
                 
                 # h. Read data
                 logger.debug(f"Reading EOG data for {sub}, {ses}, {','.join(eog_chan)}")
                 try:
                    segments = fetch(dset, annot, cat = (0,0,1,1), stage = stage, 
                                     cycle = cycle, reject_epoch = True, 
                                     reject_artf = self.reject)
                    segments.read_data(eog_chan, ref_chan, 
                                       grp_name = self.grp_name)
                 except Exception as error:
                    logger.error(error.args[0])
                    logger.warning(f"Problem loading channel {str(eog_chan)} for {sub}, {ses} ... ")
                    flag+=1
                    continue
                 
                 events_all = DataFrame(data=None)
                 for seg in segments:
                     loc = seg['data'].data[0][0]
                     roc = seg['data'].data[0][1]
                     rem = rem_detect(roc, loc, sf = dset.header['s_freq'], 
                                      hypno = None, include = 4, amplitude = amplitude, 
                                      duration = duration, freq_rem = (0.5, 5), 
                                      remove_outliers = False, verbose = False)
                     
                     events = rem.summary()
                     
                     # Get start of REM segment 
                     seg_start = seg['data'].axis['time'][0][0]
                     
                     rem_evts = []
                     
                     for e in events.index:
                         rem_evts.append({'name':'rem',
                                          'start': seg_start + events.loc[e]['Start'],
                                          'end': seg_start + events.loc[e]['End'],
                                          'chan': [''],
                                          'stage': '', 
                                          'quality': 'Good', 
                                          'cycle': ''})
                         
                         # Update summary with real start and end times too.
                         events.loc[e, 'Start'] = seg_start + events.loc[e]['Start']
                         events.loc[e, 'End'] = seg_start + events.loc[e]['End']
                         
                     grapho = graphoelement.Graphoelement()
                     grapho.events = rem_evts       
                     grapho.to_annot(annot, 'rem')
                     concat((events_all, events))
                     
                 #Save summary to dataset
                 events_all.to_csv(f'{backup}/{sub}_{ses}_rems_summary.csv')
                 
                 print('HOLD')






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  9 17:38:31 2022

@author: nathancross
"""

from os import listdir, mkdir, path
import shutil
from wonambi.attr import Annotations


def extract_grouped_markers(xml_dir, out_dir, part, visit, rater, chan, 
                            target, splitter, out_event, file_type ='edf'):
    
    '''
     Inserts virtual markers into Annotations file 
    '''              
    
    # Make output directory
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    # Loop through records
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not(p.startswith('.'))]
    else:
        print('')
        print("ERROR: 'part' must either be an array of subject ids or = 'all' ")
        print('')
    
    for i, p in enumerate(part):
        
        if not path.exists(out_dir + '/' + p):
            mkdir(out_dir + '/' + p)
        
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not(x.startswith('.'))]
        for j, vis in enumerate(visit): 
            if not path.exists(xml_dir + '/' + p + '/' + vis + '/'):
                print(f'WARNING: input folder missing for Subject {p}, visit {j}, skipping..')
                continue
            else:
                if not path.exists(out_dir + '/' + p + '/' + vis):
                    mkdir(out_dir + '/' + p + '/' + vis)
                xml_file = [x for x in listdir(xml_dir + '/' + p + '/' + vis) if x.endswith('.xml') if not x.startswith(".")] 
                
                if len(xml_file) == 0:
                    print(f'WARNING: annotations does not exist for Subject {p}, visit {j} - check this. Skipping..')
                elif len(xml_file) >1:
                    print(f'WARNING: multiple annotations files exist for Subject {p}, visit {j} - check this. Skipping..')
                else:
                    print(f'Creating virtual markers for Subject {p}, Visit {vis}')
                    
                    print(f'Grouping markers for subject {p}, visit {vis}')
                    
                    # Copy annotations file before beginning
                    xdir = xml_dir + '/' + p + '/' + vis + '/' 
                    odir = out_dir + p + '/' + vis + '/'
                    backup_file = (f'{odir}{xml_file[0]}')
                    shutil.copy(xdir + xml_file[0], backup_file)
                    
                    # Open recording and annotations files
                    annot = Annotations(backup_file, rater_name=rater)  
                    
                    # Get list of events for target and splitters
                    a = annot.get_events(target)
                    b = annot.get_events(splitter)
                    
                    # Generate list of splitter start times
                    splittimes = [x['start'] for x in b]
                    
                    # Check if splitter begins before target
                    if a[0]['start'] < splittimes[0]:
                        print('WARNING: No splitter found before first target')
                    
                    #
                    out = []
                    for e1,evt1 in enumerate(a):
                       if e1 >0:
                           splitter_trgt = [x for x in b if a[e1-1]['end'] < x['start'] 
                                            < evt1['start']]
                            
                           if len(splitter_trgt) > 0:
                               out.append(evt1)
                    
     
                    for item in out:
                        item['name'] = out_event
                    
                    annot.add_events(out)
                    
                    
                    
                    
                    


"""Packages containing power spectral analyses

"""


from .audit import (check_dataset, make_bids, extract_channels, 
                    track_processing, check_fooof)
from .load import (select_input_dirs, select_output_dirs, check_chans,
                   load_channels, rename_channels, check_adap_bands,
                   read_manual_peaks, load_adap_bands, read_inversion)
from .logs import (CustomFormatter, CustomLogfileFormatter, create_logger,
                   create_logger_empty, create_logger_basic, 
                   create_logger_outfile)
from .process import parallel
from .misc import (clean_annots, remove_event, remove_duplicate_evts, 
                   remove_duplicate_evts, merge_xmls, rainbow_merge_evts, 
                   rename_evts, laplacian_mne, notch_mne, notch_mne2, 
                   bandpass_mne, csv_stage_import)
from .splitter import extract_grouped_markers




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Created on Mon Mar 27 14:18:54 2023

@author: nathancross
"""

" Spectrogram on events."

from copy import deepcopy
from datetime import datetime, date
from numpy import (append, flip, mean, where, zeros)
from os import listdir, mkdir, path, walk
from pandas import DataFrame
from safepickle import dump, load
from scipy.signal import spectrogram
from scipy.stats import binned_statistic
import sys
from ..utils.misc import laplacian_mne, notch_mne, notch_mne2
from ..utils.logs import create_logger, create_logger_outfile
from ..utils.load import load_channels, rename_channels

from wonambi.attr import Annotations 
from wonambi.trans import fetch
from wonambi import Dataset
import shutil

def event_spectrogram(self, in_dir, xml_dir, out_dir, subs, sessions, stage, 
                      cycle_idx, chan, ref_chan, rater, grp_name, evt_type, buffer, 
                      invert, cat, filter_opts, outfile=False, filetype = '.edf', 
                      progress=True, tracking = None):

        
    ### 0.a Set up logging
    flag = 0
    tracking = self.tracking
    if outfile == True:
        today = date.today().strftime("%Y%m%d")
        now = datetime.now().strftime("%H:%M:%S")
        logfile = f'{self.log_dir}/detect_spindles_Spectrogram_{today}_log.txt'
        logger = create_logger_outfile(logfile=logfile, name='Detect spindles')
        logger.info('')
        logger.info(f"-------------- New call of 'Spectrogram' evoked at {now} --------------")
    elif outfile:
        logfile = f'{self.log_dir}/{outfile}'
        logger = create_logger_outfile(logfile=logfile, name='Detect spindles')
    else:
        logger = create_logger('Detect spindles')
    
    logger.debug(r"""    Computing spectrogram...
          
                 
             |................................................
             |........................ ..  ...................
             |...............'.......    .....................
             |....  ......... .......  .......'...............
             |....   ..'..... .......  '......'......'........
             |...       ..... .......   ......'......'........
             |...    .. ..... ................'......'........
             |...  .... ..... ............... '......'........
             |...  ...; .... :......'........  ...... ........
             |... ....   •.,    •... '.......  '..... ........
             |.......;   '.;  . .... '......    ..... ........
             |......'     '.  . '...   ...       ..,  '.......
             |________________________________________________
                     (Hz)
          
          """)
    
    # Make output directory
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    # b. Check input list
    subs = self.subs
    if isinstance(subs, list):
        None
    elif subs == 'all':
            subs = listdir(self.rec_dir)
            subs = [p for p in subs if not '.' in p]
    else:
        logger.error("'subs' must either be an array of subject ids or = 'all' ") 
    
    
    # a. Begin loop through participants
    subs.sort()
    for i, sub in enumerate(subs):
        tracking[f'{sub}'] = {}
        # b. Begin loop through sessions
        sessions = self.sessions
        if sessions == 'all':
            sessions = listdir(self.rec_dir + '/' + sub)
            sessions = [x for x in sessions if not '.' in x]   
        
        for v, ses in enumerate(sessions):
            logger.info('')
            logger.debug(f'Commencing {sub}, {ses}')
            tracking[f'{sub}'][f'{ses}'] = {'spectrogram':{}} 

            ## c. Load recording
            rdir = self.rec_dir + '/' + sub + '/' + ses + '/eeg/'
            try:
                edf_file = [x for x in listdir(rdir) if x.endswith(filetype)]
                dset = Dataset(rdir + edf_file[0])
            except:
                logger.warning(f' No input {filetype} file in {rdir}')
                break
    
            ## d. Load annotations
            xdir = self.xml_dir + '/' + sub + '/' + ses + '/'
            try:
                xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]
                # Copy annotations file before beginning
                if not path.exists(self.out_dir):
                    mkdir(self.out_dir)
                if not path.exists(self.out_dir + '/' + sub):
                    mkdir(self.out_dir + '/' + sub)
                if not path.exists(self.out_dir + '/' + sub + '/' + ses):
                    mkdir(self.out_dir + '/' + sub + '/' + ses)
                backup = self.out_dir + '/' + sub + '/' + ses + '/'
                backup_file = (f'{backup}{sub}_{ses}_spindle.xml')
                if not path.exists(backup_file):
                    shutil.copy(xdir + xml_file[0], backup_file)
                else:
                    logger.debug(f'Annotations file already exists for {sub}, {ses}, any previously detected events will be overwritten.')
                annot = Annotations(backup_file, rater_name=rater)
            except:
                logger.warning(f' No input annotations file in {xdir}')
                break
            
            ## e. Get sleep cycles (if any)
            if cycle_idx is not None:
                all_cycles = annot.get_cycles()
                cycle = [all_cycles[y - 1] for y in cycle_idx if y <= len(all_cycles)]
            else:
                cycle = None
            
            ## f. Channel setup 
            pflag = deepcopy(flag)
            flag, chanset = load_channels(sub, ses, self.chan, self.ref_chan,
                                          flag, logger)
            if flag - pflag > 0:
                logger.warning(f'Skipping {sub}, {ses}...')
                break
            
            newchans = rename_channels(sub, ses, self.chan, logger)
            
            for c, ch in enumerate(chanset):
                
                # 5 Channel stuff
                # a. Rename channel for output file (if required)
                if newchans:
                    fnamechan = newchans[ch]
                else:
                    fnamechan = ch
                # b. Set ref channel name for log output
                if not chanset[ch]:
                    logchan = ['(no re-refrencing)']
                else:
                    logchan = chanset[ch]
                    
                if filter_opts['laplacian']:
                    chans = filter_opts['lapchan']
                else:
                    chans = [ch]
                    ref_chan=chanset[chans]

                # h. Read data
                logger.debug(f"Reading EEG data for {sub}, {ses}, {str(ch)}:{'-'.join(logchan)}")
                logger.debug(f"Using filter settings:\nNotch filter: {filter_opts['notch']}")
                logger.debug(f"Notch harmonics filter: {filter_opts['notch_harmonics']}")
                logger.debug(f"Laplacian filter: {filter_opts['laplacian']}")

                
                try:
                    segments = fetch(dset, annot, cat=cat, stage=self.stage, 
                                     cycle=cycle, evt_type=evt_type,
                                     buffer=buffer, reject_artf=self.reject)

                    segments.read_data(chans, ref_chan, grp_name=self.grp_name)
                except Exception as error:
                    logger.error(type(error).__name__, "–", error)
                    flag+=1
                    break
                
                if len(segments) <1:
                    logger.error(f'No events found for {sub}, {ses}, {str(ch)}, skipping...')
                    continue
                
                nsegs=[]
                # Check concatenation of stages
                if cat[0] == 0 and cat[1] == 0:
                    logger.debug('Splitting cycles and stages')
                elif cat[0] == 0:
                    logger.debug('Splitting cycles')
                    for cy in cycle:
                        segs = [s for s in segments if cy in s['cycle']]
                        nsegs.append(segs)
                elif cat[1] == 0:
                    logger.debug('Splitting stages')
                    for st in stage:
                        segs = [s for s in segments if st in s['stage']]
                        nsegs.append(segs)
                else:
                    nsegs = [segments]
                




                
    # Check input visits
    if isinstance(sessions, list):
        None
    elif sessions == 'all':
        lenvis = set([len(next(walk(in_dir + x))[1]) for x in subs])
        if len(lenvis) > 1:
            print(colored('WARNING |', 'yellow', attrs=['bold']),
                  colored('number of visits are not the same for all subjects.',
                          'white', attrs=['bold']))
            print('')
    else:
        print('')
        print(colored('ERROR |', 'red', attrs=['bold']),
              colored("'visit' must either be an array of subject ids or = 'visit'" ,
                      'cyan', attrs=['bold']))
        print('')
    
    # Ensure correct concatenation setup
    cat = list(cat)
    cat[2] = 0 #enforce non-concatenation of signal
    cat[3] = 0 #enforce non-concatenation of event types
    cat = tuple(cat)
    
    # Loop through participants and visits
    part.sort()
    for i, p in enumerate(part):
        
        if not path.exists(out_dir + '/' + p):
            mkdir(out_dir + '/' + p)
        
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not '.' in x]
        visit.sort()    
        for j, vis in enumerate(visit): 
            if not path.exists(xml_dir + '/' + p + '/' + vis + '/'):
                print(colored('WARNING |', 'yellow', attrs=['bold']),
                      colored(f'input folder missing for Subject {p}, visit {j} ({vis}), skipping...',
                              'white', attrs=['bold']))
                print('')
                continue
            else:
                
                if not path.exists(out_dir + '/' + p + '/' + vis):
                    mkdir(out_dir + '/' + p + '/' + vis)
                rec_file = [s for s in listdir(rec_dir + '/' + p + '/' + vis) if 
                            ('.edf') in s or ('.rec') in s or ('.eeg')  in s if not s.startswith(".")]
                xml_file = [x for x in listdir(xml_dir + '/' + p + '/' + vis) if 
                            x.endswith('.xml') if not x.startswith(".")] 
                
                # Open recording and annotations files (if existing)
                if len(xml_file) == 0:
                    print(colored('WARNING |', 'yellow', attrs=['bold']),
                          colored(f'annotations does not exist for Subject {p}, visit {j} ({vis}) - check this. Skipping...',
                                  'white', attrs=['bold']))
                    print('')
                elif len(xml_file) >1:
                    print(colored('WARNING |', 'yellow', attrs=['bold']),
                          colored(f'multiple annotations files exist for Subject {p}, visit {j} ({vis}) - check this. Skipping...',
                                  'white', attrs=['bold']))
                    print('')
                else:
                    
                    ###########################            DEBUGGING              ###########################
                    with open(out_dir + f'/debug_{p}.txt', 'w') as f:
                        f.write('opening participant edf and xml')
                    ###########################            DEBUGGING              ###########################
    
                # Load data
                dset = Dataset(rec_dir + '/' + p + '/' + vis + '/' + rec_file[0]) 
                s_freq = dset.header['s_freq']
                annot = Annotations(xml_dir + '/' + p + '/' + vis + '/' + xml_file[0], 
                                    rater_name=None)
                
                # Get sleep cycles
                if cycle_idx is not None and cat[0] == 1:
                    all_cycles = annot.get_cycles()
                    scycle = [all_cycles[i - 1] for i in cycle_idx if i <= len(all_cycles)]
                else:
                    scycle = [None]
                
                # Loop through channels
                for k, ch in enumerate(chan):
                    
                    print(f'Reading data for {p}, visit {vis}, channel {ch}')
                    chan_full = ch + ' (' + grp_name + ')'
                    
                    
                    # Loop through sleep cycles 
                    for l, cyc in enumerate(scycle):
                        print('')
                        # Select and read data
                        if cycle_idx is not None:
                            print(f'Analysing, cycle {l+1}')
                        else:
                            print('Analysing, all cycles')
                        print('')
                        print('Using filter settings:')
                        print('')
                        print(colored('Notch filter:','white', attrs=['bold']),
                              colored(f"{filter_opts['notch']}", 'yellow', attrs=['bold']))
                        print(colored('Notch harmonics filter:','white', attrs=['bold']),
                              colored(f"{filter_opts['notch_harmonics']}", 'yellow', attrs=['bold']))
                        print(colored('Laplacian filter:','white', attrs=['bold']),
                              colored(f"{filter_opts['laplacian']}", 'yellow', attrs=['bold']))
                        
                        # Fetch segments from recording
                        segments = fetch(dset, annot, cat=cat, chan_full=[chan_full], 
                                         cycle=cyc, evt_type=evt_type, stage=stage,
                                         buffer=buffer)
                        if filter_opts['laplacian'] or filter_opts['notch'] or filter_opts['notch_harmonics']:
                            chans = filter_opts['lapchan']
                        else:
                            chans = [ch]
                        segments.read_data(chan=chans, ref_chan=ref_chan)
                        
                        if len(segments) <1:
                            print(colored('WARNING |', 'yellow', attrs=['bold']),
                                  colored('No segments found.',
                                          'white', attrs=['bold']))
                            
                        nsegs=[]
                        # Check concatenation of stages
                        if cat[1] == 0:
                            print('Splitting stages')
                            for s, st in enumerate(stage):
                                segs = [s for s in segments if st in s['stage']]
                                nsegs.append(segs)
                        else:
                            nsegs = [segments]
                        
                        for s in range(len(nsegs)):
                            segments = nsegs[s]
                            
                            if cat[1] == 1:
                                stagename = ''.join(stage) 
                            else: 
                                stagename = stage[s]
                                
                            print('')
                            print(f'Stage {stagename}')
                            print('Creating spectrogram')
                            print(f'No. Segments = {len(segments)}')
                           
                            z=0
                            for m, seg in enumerate(segments):
                                
                                # Print out progress
                                if progress:
                                    z +=1
                                    j = z/len(segments)
                                    sys.stdout.write('\r')
                                    sys.stdout.write(f"Progress: [{'=' * int(50 * j):{50}s}] {int(100 * j)}%")
                                    sys.stdout.flush()
                                
                                # Select data from segment
                                data = seg['data']
                                
                                # Find sampling frequency
                                s_freq = data.s_freq
    
                                # Apply filtering (if necessary)
                                if filter_opts['notch']:
                                    selectchans = list(data.chan[0])
                                    data.data[0] = notch_mne(data, oREF=filter_opts['oREF'], 
                                                                channel=selectchans, 
                                                                freq=filter_opts['notch_freq'],
                                                                rename=filter_opts['chan_rename'],
                                                                renames=filter_opts['renames'])
                                
                                if filter_opts['notch_harmonics']: 
                                    selectchans = list(data.chan[0])
                                    data.data[0] = notch_mne2(data, oREF=filter_opts['oREF'], 
                                                              channel=selectchans,
                                                              rename=filter_opts['chan_rename'],
                                                              renames=filter_opts['renames'])
                                
                                if filter_opts['laplacian']:
                                    data = laplacian_mne(data, oREF=filter_opts['oREF'], channel=ch, 
                                                         ref_chan=ref_chan, 
                                                         laplacian_rename=filter_opts['laplacian_rename'], 
                                                         renames=filter_opts['renames'])
                                    dat = data[0]
                                else:
                                    dat = data()[0][0]
                                
                                
                                # Check polarity of recording
                                if isinstance(polar, list):
                                    polarity = polar[i]
                                else:
                                    polarity = polar
                                if polarity == 'opposite':
                                    dat = dat*-1 
                        
                                if m == 0:
                                    dat = dat[int((len(dat)/2)-(s_freq*buffer)):int((len(dat)/2)+(s_freq*buffer))]
                                    out = dat.reshape((1,len(dat)))
                                else:
                                    dat = dat[int((len(dat)/2)-(s_freq*buffer)):int((len(dat)/2)+(s_freq*buffer))]
                                    dat = dat.reshape((1,len(dat)))
                                    out = append(out,dat,axis=0)
                                     
                        
                        # Calaculate spectrogram
                        f,t,spect = spectrogram(out, fs=s_freq, nfft=filter_opts['nfft'], 
                                                noverlap=filter_opts['noverlap'],)
                        
                        # Apply mask to extract only frequencies you want
                        freqs = filter_opts['bandpass']
                        highpass = where(f==freqs[0])[0][0]
                        lowpass = where(f==freqs[1])[0][0]
                        spect = flip(spect[:,highpass:lowpass,:], axis=1)
                        f = f[highpass:lowpass]
                        
                        # Prepare filename
                        if cat[0] == 1:
                            cyclename = 'wholenight'
                        else: 
                            cyclename = f'cycle{l+1}' 
                            
                        # Save average spectrogram to dataframe
                        spectav = mean(spect,axis=0)
                        d = DataFrame(spectav)
                        d.columns = t
                        d.index = flip(f)
                        d.to_csv(path_or_buf=out_dir + '/' + p + '/' + vis + '/' + 
                                 p + '_' + vis + '_' + ch + '_' + stagename +  '_' + cyclename + 
                                 '_spectrogram.csv', sep=',')
                        
                        # Save spectrogram for every event to pickle file
                        spectro = {'spect':spect, 'f':f, 't':t}
                        with open(out_dir + '/' + p + '/' + vis + '/' + 
                                 p + '_' + vis + '_' + ch + '_' + stagename + '_' + cyclename + 
                                  '_spectrogram.p', 'wb') as ff:
                             dump(spectro, ff)
                             
                        # Save raw and averaged event signal to pickle file
                        datav = mean(out, axis=0)
                        signal = {'raw':out, 'average':datav}
                        with open(out_dir + '/' + p + '/' + vis + '/' + 
                                 p + '_' + vis + '_' + ch + '_' + stagename + '_' + cyclename + 
                                  '_signal.p', 'wb') as ff:
                             dump(signal, ff)
    
    print('The function event_spectrogram completed without error.')
    return
                
                
                
                

def event_spectrogram_grouplevel(in_dir, out_dir, part, visit, chan, stage, cat, 
                                 cycle_idx):                        
             
    '''
    This script combines the output from the function event_spectrogram, and formats it
    in a group-level dataframe for statistical analyses.
    The outputs provided by this script will be, for each visit and EEG channel:
        1. A pickle file (per channel and visit) with the night-average spectrogram for
            every subject and the night-average signal of each event.
        2. A .csv file (per channel and visit) containing the group averaged spectrogram.
        
    '''   
    
    # Make output directory
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    ## BIDS CHECKING
    # Check input participants
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(in_dir)
            part = [ p for p in part if not '.' in p]
    else:
        print('')
        print(colored('ERROR |', 'red', attrs=['bold']),
              colored("'part' must either be an array of subject ids or = 'all'" ,
                      'cyan', attrs=['bold']))
        print('')
        
    # Check input visits
    if isinstance(visit, list):
        None
    elif visit == 'all':
        lenvis = set([len(next(walk(in_dir + x))[1]) for x in part])
        if len(lenvis) > 1:
            print(colored('WARNING |', 'yellow', attrs=['bold']),
                  colored('number of visits are not the same for all subjects.',
                          'white', attrs=['bold']))
            print('')
        visit = list(set([y for x in part for y in listdir(in_dir + x)  if '.' not in y]))
    else:
        print('')
        print(colored('ERROR |', 'red', attrs=['bold']),
              colored("'visit' must either be an array of subject ids or = 'visit'" ,
                      'cyan', attrs=['bold']))
        print('')
            
    # Create output dataframe
    if cycle_idx is not None:
        all_ampbin = zeros((7, len(part), 6), dtype='object')
    else:
        all_ampbin = zeros((7, len(part)), dtype='object')
    
    # Check for stage setup
    if cat[1] == 1:
        stage = [''.join(stage) ]

    for st, stagename in enumerate(stage): # Loop through stages
        for k, ch in enumerate(chan):      # Loop through channels
            print('')
            print(f'CHANNEL {ch}')
            
            
            for j, vis in enumerate(visit): 
                z=0
                index=[]
                part.sort()
                spectgrp = zeros((len(part)), dtype='object')
                signalgrp = zeros((len(part)), dtype='object')
                for i, p in enumerate(part):    # Loop through participants
                    index.append(p)
                    if not path.exists(in_dir + '/' + p + '/' + vis + '/'):
                        print(colored('WARNING |', 'yellow', attrs=['bold']),
                              colored(f'input folder missing for Subject {p}, visit {vis}, skipping..',
                                      'white', attrs=['bold']))
                        continue
                    else:
                        
                        # Spectrogram
                        # Define pickle files for spectrogram
                        p_files = [s for s in listdir(in_dir + '/' + p + '/' + vis) 
                                   if '_spectrogram.p' in s] 
                        p_files = [s for s in p_files if ch in s]
                        p_files = [s for s in p_files if '_'+stagename+'_' in s]
                        
                        # Open files containing spectrogram (if existing)
                        if len(p_files) == 0:
                            print(colored('WARNING |', 'yellow', attrs=['bold']),
                                  colored(f'spectrogram does not exist for Subject {p}, visit {vis}, stage {stagename}, channel {ch}. Skipping..',
                                          'white', attrs=['bold']))
                        elif len(p_files)>1:
                            print(colored('WARNING |', 'yellow', attrs=['bold']),
                                  colored(f'multiple spectrogram files exist for Subject {p}, visit {vis}, stage {stagename}, channel {ch}. Skipping..',
                                          'white', attrs=['bold']))
                            print('')
                        else:
                            print(f'Extracting spectrogram for ... Subject {p}, visit {vis}')
                            pfile = in_dir + '/' + p + '/' + vis + '/' + p_files[0]
                            with open(pfile, 'rb') as ff:
                                pick = load(ff)
                            
                            spect = pick['spect']
                            f = pick['f']
                            t = pick['t']
                            spectgrp[i] = mean(spect, axis=0)
                            
                        ## Raw signal  
                        # Define pickle files for raw signal
                        p_files = [s for s in listdir(in_dir + '/' + p + '/' + vis) 
                                   if '_signal.p' in s] 
                        p_files = [s for s in p_files if ch in s]
                        p_files = [s for s in p_files if '_'+stagename+'_' in s]
                        
                        # Open files containing event signal (if existing)
                        if len(p_files) == 0:
                            print(colored('WARNING |', 'yellow', attrs=['bold']),
                                  colored(f'signal file does not exist for Subject {p}, visit {vis}, stage {stagename}, channel {ch}. Skipping..',
                                          'white', attrs=['bold']))
                        elif len(p_files)>1:
                            print(colored('WARNING |', 'yellow', attrs=['bold']),
                                  colored(f'multiple signal files exist for Subject {p}, visit {vis}, stage {stagename}, channel {ch}. Skipping..',
                                          'white', attrs=['bold']))
                            print('')
                        else:
                            print(f'Extracting raw signal for ... Subject {p}, visit {vis}')
                            pfile = in_dir + '/' + p + '/' + vis + '/' + p_files[0]
                            with open(pfile, 'rb') as ff:
                                pick = load(ff)
                            
                            signal = pick['average']
                            signalgrp[i] = binned_statistic(range(0,len(signal)), signal,  
                                                            bins=spectgrp[i].shape[1])[0]
                            
                    
                # Save all spectrograms and event signal to pickle file
                grouped_spectrogram = {'spectrogram':spectgrp, 'signal':signalgrp}
                with open(out_dir + ch + '_' + stagename + '_visit_' + vis +
                          '_group_spectrogram.p', 'wb') as ff:
                     dump(grouped_spectrogram, ff)
                 
                    
                # Save average spectrogram to dataframe
                spectav = mean(spectgrp,axis=0)
                d = DataFrame(spectav)
                d.columns = t
                d.index = flip(f)
                d.to_csv(path_or_buf=out_dir + ch + '_' + stagename +  '_visit_' + vis + 
                         '_spectrogram.csv', sep=',')  
                            
    return
                
                
                
                





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 31 16:17:18 2024

@author: m.ela
"""

#%%
" Opening a recording in Wonambi."

# 1. Annotations file

# a. load function
from wonambi.attr import Annotations 
import shutil
xml = '/Users/m.ela/Desktop/DATA_temp/QC_ROCK_temp/ROCK25/STIM1/ROCK25_E_STIM1_scores_LB_AP.xml'

new_xml = '/Users/m.ela/Desktop/DATA_temp/QC_ROCK_temp/OUT/marker/ROCK25_E_STIM1_scores_test.xml'

shutil.copy(xml, new_xml)

# b. load annotations file
annot = Annotations(new_xml)

events = annot.get_events()  ## get all events in a list
#events = annot.get_events(name = 'Staresina2015', chan = 'Cz') #get event named X on channel X


# 2. EDF
# a. load function
from wonambi import Dataset
edf_file = '/Users/m.ela/Desktop/DATA_temp/QC_ROCK_temp/ROCK25/STIM1/ROCK25_E_STIM1.edf'

# b. Load edf in Wonambi
dset = Dataset(edf_file)
#header = dset.header #to read header with info on .edf
s_freq = dset.header['s_freq']

# c. VERSION1: Read data 
    # i. (between specific timepoints)
data_ref = dset.read_data(chan=['AUX'], begtime=0, endtime=3000) #extract data from channel of interest from time x to timey
# it will create de Chan * Time datatype with a weird array within array 
#dir(data_ref) #to see the header 
#data_ref.data #to see the actually data (512pt per second)
data = data_ref.data[0][0] # 2 brackes to have data in correct shape (that can be visualize)
    #ii. visualize raw
import matplotlib.pyplot as plt
plt.plot(data)



#d. VERSION 2: load in segments
from wonambi.trans import fetch
cat = (1,1,1,1)
chans = ['AUX']
ref_chan = ['M1','M2']
stage = ['NREM2','NREM3']
chan_full = ['AUX (aux)']
evt_type = None
segments = fetch(dset, annot, cat=cat, chan_full=chan_full, 
                 cycle=None, evt_type=evt_type, stage=stage,
                 buffer=3)
    
segments.read_data(chan=chans) # this makes segments into a list of segments

seg = segments[0] # call the first segment, which is a dictionary of metadata and data
seg['data'].data[0][0] # Just like in VERSION1, # 2 brackes to have data in correct shape (that can be visualize)



# 3. Filtering
from numpy import roll
from scipy.signal import firwin, kaiserord, lfilter

#dat = seg['data'].data[0][0].T*-1 # in case you want to invert the signal
#dat = seg['data'].data[0][0]
dat = data

# Filter (w FIR)
bandwidth = (0.25, 0.5)
dat_orig = dat
nyq_rate = s_freq / 2.0
width = 5.0/nyq_rate
N, beta = kaiserord(60.0, width)
cutoff = bandwidth[1]       # for bandpass, if you want lowpass, set cutoff = bandwidth[1]
b = firwin(N, cutoff/nyq_rate, window=('kaiser', beta))
fir = lfilter(b, 1.0, dat_orig)
delay = int(-0.5 * (N-1))
dat_filt = roll(fir, delay)


# 3. Visualise data
plt.plot(dat)
plt.plot(dat_filt)


from scipy.signal import find_peaks

peaks, _ = find_peaks(dat_filt, height=550)


evt_name = 'bed'

evts = []
for i, pk in enumerate(peaks):
    if pk - peaks[i-1] > s_freq:
        pk_sec = pk/s_freq
        evt = {'chan':['AUX (aux)'],
               'cycle':'',
               'end': pk_sec+0.1,
               'name': evt_name,
               'quality': 'Good',
               'stage': '',
               'start': pk_sec-0.1
               }
        evts.append(evt)

from wonambi.graphoelement import Graphoelement

grapho = Graphoelement()
grapho.events = evts
grapho.to_annot(annot, evt_name)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan 19 10:43:09 2024

@author: ncro8394
"""

from multiprocessing import Pool


def parallel(FUNC, cores, args):

    pool = Pool(cores)
    results = pool.starmap(FUNC, args)
    
    return results


# -*- coding: utf-8 -*-
"""
Created on Tue Oct  5 15:51:24 2021

@author: Nathan Cross
"""


from os import listdir, mkdir, path
from . cfc_func import circ_wwtest, mean_amp
from seapipe.utils.misc import bandpass_mne, laplacian_mne, notch_mne, notch_mne2
from copy import deepcopy
import shutil
from math import degrees
from numpy import (argmax, array, asarray, ceil, concatenate, histogram, interp, 
                   isnan, linspace, mean, median, nan, nanmean, ones, pi, random, 
                   reshape, save, sin, squeeze, sum, zeros) 
from pandas import DataFrame, read_csv
from safepickle import load
from pingouin import (circ_mean, circ_r, circ_rayleigh, circ_corrcc)
from scipy.stats import zscore
import sys
from tensorpac import Pac
from wonambi import Dataset
from wonambi.trans import fetch
from wonambi.attr import Annotations 
from seapipe.utils.logs import create_logger
from ..utils.load import (load_channels, load_adap_bands, rename_channels, 
                          load_sessions, read_inversion, read_manual_peaks)
from ..utils.misc import infer_polarity


def pac_method(method, surrogate, correction, list_methods=False):
    
    ''' Formats the method and corrections to be applied through Tensorpac:
        https://etiennecmb.github.io/tensorpac/auto_examples/index.html#tutorials
    '''
    # Calculate Coupling Strength (idpac)
    methods = {1: 'Mean Vector Length (MVL) [Canolty et al. 2006 (Science)]',
               2 : 'Modulation Index (MI) [Tort 2010 (J Neurophys.)]',
               3 : 'Heights Ratio (HR) [Lakatos 2005 (J Neurophys.)]',
               4 : 'ndPAC [Ozkurt 2012 (IEEE)]',
               5 : 'Phase-Locking Value (PLV) [Penny 2008 (J. Neuro. Meth.), Lachaux 1999 (HBM)]',
               6 : 'Gaussian Copula PAC (GCPAC) `Ince 2017 (HBM)`'}
    surrogates = {0 :'No surrogates', 
                  1 : 'Swap phase / amplitude across trials [Tort 2010 (J Neurophys.)]',
                  2 : 'Swap amplitude time blocks [Bahramisharif 2013 (J. Neurosci.) ]',
                  3 : 'Time lag [Canolty et al. 2006 (Science)]'}
    corrections = {0 : 'No normalization',
                   1 : 'Subtract the mean of surrogates',
                   2 : 'Divide by the mean of surrogates',
                   3 : 'Subtract then divide by the mean of surrogates',
                   4 : 'Z-score'}
    
    if list_methods:
        idpac = [methods] + [surrogates] + [corrections]
    else:
        meth = [x for x in methods if method in methods[x]][0]
        surr = [x for x in surrogates if surrogate in surrogates[x]][0]
        corr = [x for x in corrections if correction in corrections[x]][0]
        idpac = tuple((meth,surr,corr))
    
    return idpac


class octopus:

    def __init__(self, rootpath, rec_dir, xml_dir, out_dir, chan, ref_chan, 
                 grp_name, stage, rater = None, subs = 'all', 
                 sessions = 'all', reject_artf = ['Artefact', 'Arou', 'Arousal'], 
                 tracking = None):
        
        self.rootpath = rootpath
        self.rec_dir = rec_dir
        self.xml_dir = xml_dir
        self.out_dir = out_dir
        self.chan = chan
        self.ref_chan = ref_chan
        self.grp_name = grp_name
        self.stage = stage
        self.rater = rater
        self.reject = reject_artf
        
        self.subs = subs
        self.sessions = sessions
        
        if tracking == None:
            tracking = {'event_pac':{}}
        self.tracking = tracking


    def pac_it(self, cycle_idx, cat, nbins, filter_opts, 
                 epoch_opts, frequency_opts, event_opts, filetype = '.edf',
                 idpac = (2,3,4), evt_type = None, min_dur = 1, 
                 adap_bands_phase = 'Fixed', frequency_phase = (0.5,1.25), 
                 adap_bands_amplitude = 'Fixed', frequency_amplitude = (11,16), 
                 adap_bw = 4, invert = False, progress=True,
                 logger = create_logger('Phase-amplitude coupling')):

        '''
        O.C.T.O.P.U.S
        
        Oscillatory Coupling: Timed Oscillations by Phase modUlation in Sleep 
        
        This script runs Phase Amplitude Coupling analyses on sleep EEG data. 
        The method for calculating PAC is set by the parameter <idpac>. 
        For more information on the available methods, refer to the documentation of 
        tensorpac (https://etiennecmb.github.io/tensorpac/index.html) or the article
        (Combrisson et al. 2020, PLoS Comp Bio: https://doi.org/10.1371/journal.pcbi.1008302)
        
        The script does the following:
            1. Extracts the EEG signal at each event specified by <evt_type> ± a buffer
               on either side of length (in sec) specified by <buffer>.
            2. For these EEG segments, filters the signal within a given frequency range
               specified by <fpha> to obtain the phase, and again within a given frequency 
               range specified by <famp>.  
            3. FOR EACH EACH EVENT: the instantaneous amplitude of the signal filtered 
               within a given frequency range specified by <famp> will be calculated 
               via the Hilbert transform, and the amplitude will be averaged across a 
               set number of phase bins specified by <nbins>. The phase bin with the 
               maxmimum mean amplitude will be stored.
            4. ACROSS ALL EVENTS: the average phase bin with the maximum amplitude will 
               be calculated (circular mean direction).
            5. The filtered events will also be concatenated and stored in blocks of 50,
               so that the PAC strength (method pecficied by the 1st entry in <idpac>)
               can be calculated AND surrogates can be accurately generated to test for
               the significance of PAC in each participant. The generation of surrogates
               is specified by the 2nd entry in <idpac>, and the correction of PAC 
               strength is also calculated, specified by the 3rd entry in <idpac>.
            6. Other metrics are also calculated for each participant and visit, notably:
                - mean vector length (given from mean circular calculation)
                - correlation between amplitudes (averaged over all events) and the phase 
                   giving sine wave
                - Rayleigh test for non-uniformity of circular data (sig. test for 
                                                                     preferred phase)
                
               
        If laplacian = True then a Laplacian spatial filter will be applied to remove high frequency EMG 
        noise. In this scenario you will need to provide a list of channel names to include in the laplacian
        spatial filtering. 
                        ## WARNING: TEST WHAT THE LAPLACIAN FILTER DOES TO THE POWER SPECTRUM BEFORE USING
                                    THIS OPTION
        
        If adap_bands_phase = True then the phase signal will be filtered within an adapted 
        frequency range for each individual subject or recording.
        
        If adap_bands_amplitude = True then the amplitude signal will be filtered within an adapted 
        frequency range for each individual subject or recording.
        
        The output provided by this script will be an array of size:
            [#cycles x #bins] (if cycle_idx is not None)
            [1 x #nbins]      (if cycle_idx is None)
                                            - corresponding to the mean amplitude of the signal 
                                             (across all cycles or events) per phase bin.
    
        '''
        # Get method descriptions
        pac_list = pac_method(0, 0, 0, list_methods=True)
        methods = pac_list[0]
        surrogates = pac_list[1]
        corrections = pac_list[2]
    
        ### 0.a Set up logging
        tracking = self.tracking
        flag = 0
        
        
        logger.info('')
        logger.debug(rf"""Commencing phase-amplitude coupling pipeline... 
                                
                                        ___
                                     .-'   `'.
                                    /         \
                                    |         ;
                                    |         |           ___.--,
                           _.._     |0) ~ (0) |    _.---'`__.-( (_.
                     __.--'`_..'.__.\    '--. \_.-' ,.--'`     `""`
                    ( ,.--'`   ',__ /./;   ;, '.__.'`    __
                  _`) )    .---.__.' / |   |\   \__..--""  ""--.,_
                 `---'   .'.''-._.-'`_./  /\ '.  \ _.-~~~````~~~-._`-.__.'
                         | |  .' _.-' |  |  \  \  '.               `~---`
                         \ \/ .'     \  \   '. '-._)
                          \/ /        \  \    `=.__`~-.
                          / /\         `) )    / / `"".`\
                    , _.-'.'\ \        / /    ( (     / /
                    `--~`   ) )    .-'.'      '.'.  | (
                            (/`    ( (`          ) )  '-;
                            `      '-;         (-'
    
                
                Oscillatory Coupling: Timed Oscillations by Phase modUlation in Sleep
                (O.C.T.O.P.U.S)
                
                Method: {methods[idpac[0]]}
                Correction: {surrogates[idpac[1]]}
                Normalisation: {corrections[idpac[2]]}
                                  
                                                    """,)
        ### 0.b. Set up organisation of export
        if cat[0] + cat[1] == 2:
            model = 'whole_night'
            logger.debug('Analysing PAC for the whole night.')
        elif cat[0] + cat[1] == 0:
            model = 'stage*cycle'
            logger.debug('Analysing PAC per stage and cycle separately.')
        elif cat[0] == 0:
            model = 'per_cycle'
            logger.debug('Analysing PAC per cycle separately.')
        elif cat[1] == 0:
            model = 'per_stage'  
            logger.debug('Analysing PAC per stage separately.')
        if 'cycle' in model and cycle_idx == None:
            logger.info('')
            logger.critical("To run cycles separately (i.e. cat[0] = 0), cycle_idx cannot be 'None'")
            return
        
        # Log filtering options
        logger.debug(f'Using {adap_bands_phase} bands for phase frequency.')
        logger.debug(f'Using {adap_bands_amplitude} bands for amplitude frequncy.')
        if filter_opts['notch']:
            logger.debug(f"Applying notch filtering: {filter_opts['notch_freq']} Hz")
        if filter_opts['notch_harmonics']: 
            logger.debug('Applying notch harmonics filtering.')
        if filter_opts['bandpass']:
            logger.debug(f"Applying bandpass filtering: {filter_opts['highpass']} - {filter_opts['lowpass']} Hz")
        if filter_opts['laplacian']:
            logger.debug('Applying Laplacian filtering.')
        
        # Check format of evt_type
        if not evt_type == None and not isinstance(evt_type, list):
            evt_type = [evt_type]
        
        ### 1. First we check the directories
        # a. Check for output folder, if doesn't exist, create
        if path.exists(self.out_dir):
                logger.debug("Output directory: " + self.out_dir + " exists")
        else:
            mkdir(self.out_dir)
        
        # b. Check input list
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = listdir(self.rec_dir)
                subs = [p for p in subs if not '.' in p]
        else:
            logger.error("'subs' must either be an array of subject ids or = 'all' ")   
        
        ### 2. Begin loop through dataset
        # a. Begin loop through participants
        subs.sort()
        for i, sub in enumerate(subs):
            if not sub in self.tracking['event_pac'].keys():
                self.tracking['event_pac'][sub] = {}
            # b. Begin loop through sessions
            flag, sessions = load_sessions(sub, self.sessions, self.rec_dir, flag, 
                                     logger, verbose=2) 
            
            for v, ses in enumerate(sessions):
                logger.info('')
                logger.debug(f'Commencing {sub}, {ses}')
                if not ses in self.tracking['event_pac'][sub].keys():
                    self.tracking['event_pac'][sub][ses] = {} 

                ## c. Load recording
                rdir = self.rec_dir + '/' + sub + '/' + ses + '/eeg/'
                try:
                    edf_file = [x for x in listdir(rdir) if x.endswith(filetype)]
                    dset = Dataset(rdir + edf_file[0])
                except:
                    logger.warning(f' No input {filetype} file in {rdir}')
                    flag +=1
                    break
                
                ## d. Load annotations
                xdir = self.xml_dir + '/' + sub + '/' + ses + '/'
                try:
                    xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]
                    # Copy annotations file before beginning
                    if not path.exists(self.out_dir):
                        mkdir(self.out_dir)
                    if not path.exists(self.out_dir + '/' + sub):
                        mkdir(self.out_dir + '/' + sub)
                    if not path.exists(self.out_dir + '/' + sub + '/' + ses):
                        mkdir(self.out_dir + '/' + sub + '/' + ses)
                    outpath = self.out_dir + '/' + sub + '/' + ses
                    backup_file = (f'{outpath}/{sub}_{ses}_spindle.xml')
                    if not path.exists(backup_file):
                        shutil.copy(xdir + xml_file[0], backup_file)
                    else:
                        logger.debug(f'Using annotations file from: {xdir}')
                except:
                    logger.warning(f' No input annotations file in {xdir}')
                    flag +=1
                    break
                
                # 3.a. Read annotations file
                annot = Annotations(backup_file, rater_name=self.rater)
                
                ## b. Get sleep cycles (if any)
                if cycle_idx is not None:
                    all_cycles = annot.get_cycles()
                    cycle = [all_cycles[y - 1] for y in cycle_idx if y <= len(all_cycles)]
                else:
                    cycle = None
                
                ## c. Channel setup 
                pflag = deepcopy(flag)
                flag, chanset = load_channels(sub, ses, self.chan, self.ref_chan,
                                              flag, logger)
                if flag - pflag > 0:
                    logger.warning(f'Skipping {sub}, {ses}...')
                    flag +=1
                    break
                newchans = rename_channels(sub, ses, self.chan, logger)
                
                # 4.a. Loop through channels
                for c, ch in enumerate(chanset):
                    
                    chan_full = f'{ch} ({self.grp_name})'
                    
                    # b. Rename channel for output file (if required)
                    if newchans:
                        fnamechan = newchans[ch]
                        filter_opts['renames'] = {newchans[ch]:ch}
                        filter_opts['laplacian_rename'] = True
                    else:
                        fnamechan = ch
                        filter_opts['laplacian_rename'] = False

                    if ch == '_REF':
                        filter_opts['laplacian_rename'] = False
                        if not filter_opts['oREF']:
                            try:
                                filter_opts['oREF'] = newchans[ch]
                            except:
                                logger.warning("Channel selected is '_REF' but "
                                               "no information has been given "
                                               "about what standard name applies, "
                                               "either via filter_opts['oREF'] "
                                               "or 'chanset_rename' in the tracking "
                                               "sheet (see user guide). Skipping "
                                               "channel...")
                                flag += 1
                                continue
                    else:
                        filter_opts['oREF'] = None
                        
                    # c. Set frequency bands for:
                    # Phase
                    if adap_bands_phase == 'Fixed':
                        f_pha = frequency_phase   
                    elif adap_bands_phase == 'Manual':
                        f_pha = read_manual_peaks(self.rootpath, sub, ses, ch, 
                                                 adap_bw, logger)
                    elif adap_bands_phase == 'Auto':
                        stagename = '-'.join(self.stage)
                        band_limits = f'{frequency_phase[0]}-{frequency_phase[1]}Hz'
                        f_pha = load_adap_bands(self.tracking['fooof'], sub, ses,
                                               fnamechan, stagename, band_limits, 
                                               adap_bw, logger)
                    if not f_pha:
                        logger.warning('Will use fixed frequency bands for PHASE instead.')
                        f_pha = frequency_phase
                    if not chanset[ch]:
                        logchan = ['(no re-refrencing)']
                    else:
                        logchan = chanset[ch]
                    logger.debug("Using PHASE frequency band: "
                                 f"{round(f_pha[0],2)}-{round(f_pha[1],2)} Hz "
                                 f"for {sub}, {ses}, {str(ch)}:{'-'.join(logchan)}")    
                    
                    # Amplitude
                    if adap_bands_amplitude == 'Fixed':
                        f_amp = frequency_amplitude   
                    elif adap_bands_amplitude == 'Manual':
                        f_amp = read_manual_peaks(self.rootpath, sub, ses, ch, 
                                                 adap_bw, logger)
                    elif adap_bands_amplitude == 'Auto':
                        stagename = '-'.join(self.stage)
                        band_limits = f'{frequency_amplitude[0]}-{frequency_amplitude[1]}Hz'
                        f_amp = load_adap_bands(self.tracking['fooof'], sub, ses,
                                               fnamechan, stagename, band_limits, 
                                               adap_bw, logger)
                    if not f_amp:
                        logger.warning('Will use fixed frequency bands for PHASE instead.')
                        f_amp = frequency_amplitude
                    if not chanset[ch]:
                        logchan = ['(no re-refrencing)']
                    else:
                        logchan = chanset[ch]
                    logger.debug("Using AMPLITUDE frequency band: "
                                 f"{round(f_amp[0],2)}-{round(f_amp[1],2)} Hz "
                                 f"for {sub}, {ses}, {str(ch)}:{'-'.join(logchan)}")    
                    
                    # d. Check if channel needs to be inverted for detection
                    if type(invert) == type(DataFrame()):
                        inversion = read_inversion(sub, ses, invert, ch, logger)
                        if not inversion:
                            inversion = infer_polarity(dset, annot, ch, 
                                                       chanset[ch], (1,1,1,1), 
                                                       evt_type, self.stage, 
                                                       cycle, logger)
                    elif type(invert) == bool:
                        inversion = invert
                    else:
                        logger.critical("The argument 'invert' must be set to either: "
                                        f"'True', 'False' or 'None'; but it was set as {invert}.")
                        logger.info('Check documentation for how to set up staging data:')
                        logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                        logger.info('-' * 10)
                        return
                    logger.debug(f"{'Inverting' if inversion else 'Not inverting'}"
                                 f" channel {ch} prior to detection for {sub}, {ses}")
                    
                    
                    # 5.a. Fetch data
                    logger.debug(f"Reading EEG data for {sub}, {ses}, {str(ch)}:{'-'.join(logchan)}")
                    try:
                        segments = fetch(dset, annot, cat = cat, evt_type = evt_type, 
                                     stage = self.stage, cycle = cycle, 
                                     buffer = event_opts['buffer'])
                    except Exception as error:
                        logger.error(error.args[0])
                        logger.warning(f"Skipping {sub}, {ses}, channel {str(ch)} ... ")
                        flag+=1
                        continue

                    if len(segments)==0:
                        logger.warning(f"No valid data found for {sub}, {ses}, "
                                       f"{self.stage}, Cycles:{cycle}.")
                        flag +=1
                        break
                    
                    # 5.b. Read data
                    if filter_opts['laplacian']:
                        try:
                            segments.read_data(filter_opts['lapchan'], chanset[ch]) 
                            laplace_flag = True
                        except:
                            logger.error("Channels listed in filter_opts['lapchan']: "
                                         f"{filter_opts['lapchan']} are not found "
                                         f"in recording for {sub}, {ses}.")
                            logger.warning("Laplacian filtering will NOT be run "
                                           f"for {sub}, {ses}, {ch}. Check parameters "
                                           "under: filter_opts['lapchan']")
                            segments.read_data(ch, chanset[ch])
                            laplace_flag = False
                            flag += 1
                    else:
                        segments.read_data(ch, chanset[ch])
                    
                    
                    # 6. Define PAC object
                    pac = Pac(idpac = idpac, f_pha = f_pha, f_amp = f_amp, 
                              dcomplex = filter_opts['dcomplex'], 
                              cycle = filter_opts['filtcycle'], 
                              width = filter_opts['width'], 
                              n_bins = nbins,
                              verbose='ERROR')
                    
                    # 7.a. Divide segments based on concatenation
                    nsegs=[]
                    if model == 'whole_night':
                        nsegs = [[s for s in segments]]
                        seg_label = ['whole_night']
                    elif model == 'stage*cycle':
                        seg_label = []
                        for st in self.stage:
                            for cy in cycle_idx:
                                segs = [s for s in segments if st in s['stage'] 
                                        if cy in s['cycle']]
                                nsegs.append(segs)
                                seg_label.append(f'{st}_cycle{cy}')
                    elif model == 'per_cycle':
                        seg_label = []
                        for cy in cycle_idx:
                            segs = [s for s in segments if cy in s['cycle']]
                            nsegs.append(segs)
                            seg_label.append(f'cycle{cy}')
                    elif model == 'per_stage':
                        seg_label = []
                        for st in self.stage:
                            segs = [s for s in segments if st in s['stage']]
                            nsegs.append(segs)
                            seg_label.append(f'{st}')
                    
                    # 7.b. Loop over segments and apply filtering (if required)
                    logger.info('')
                    for nsg in range(len(nsegs)):
                        seg = nsegs[nsg]
                        logger.debug(f'Analysing {seg_label[nsg]}')
                        
                        # 7.c. Create blocks
                        ampbin = zeros((len(seg), nbins))
                        ms = int(ceil(len(seg)/50))
                        longamp = zeros((ms,50),dtype=object) # initialise (blocked) amplitude series
                        longpha = zeros((ms,50),dtype=object) # initialise (blocked) phase series 
                        
                        
                        for s, sg in enumerate(seg):
                            
                            # Print out progress
                            if progress:
                                j = s/len(seg)
                                sys.stdout.write('\r')
                                sys.stdout.write("                      Progress: "
                                                 f"[{'»' * int(50 * j):{50}s}] {int(100 * j)}%")
                                sys.stdout.flush()
                            
                            out = dict(sg)
                            data = sg['data']
                            timeline = data.axis['time'][0]
                            out['start'] = timeline[0]
                            out['end'] = timeline[-1]
                            out['duration'] = len(timeline) / data.s_freq
                            if filter_opts['laplacian']:
                                selectchans = filter_opts['lapchan']
                            else:
                                selectchans = ch
                            
                            # 7.d. Notch filters
                            filtflag = 0
                            if filter_opts['notch']:
                                data.data[0], filtflag = notch_mne(data,  
                                                        channel=selectchans, 
                                                        freq=filter_opts['notch_freq'],
                                                        oREF=filter_opts['oREF'],
                                                        rename=filter_opts['laplacian_rename'],
                                                        renames=filter_opts['renames'],
                                                        montage=filter_opts['montage']
                                                        )
                                
                            if filter_opts['notch_harmonics']: 
                                data.data[0], filtflag = notch_mne2(data, 
                                                          channel=selectchans,
                                                          oREF=filter_opts['oREF'], 
                                                          rename=filter_opts['laplacian_rename'],
                                                          renames=filter_opts['renames'],
                                                          montage=filter_opts['montage']
                                                          )    
                            
                            # 7.e. Bandpass filters
                            if filter_opts['bandpass']:
                                data.data[0], filtflag = bandpass_mne(data, 
                                                            channel=selectchans, 
                                                            highpass=filter_opts['highpass'], 
                                                            lowpass=filter_opts['lowpass'], 
                                                            oREF=filter_opts['oREF'],
                                                            rename=filter_opts['laplacian_rename'],
                                                            renames=filter_opts['renames'],
                                                            montage=filter_opts['montage']
                                                            )
                            
                            # 7.f. Laplacian transform
                            if filter_opts['laplacian'] and laplace_flag:
                                data.data[0], filtflag = laplacian_mne(data, 
                                                                       channel=selectchans, 
                                                                       ref_chan=chanset[ch], 
                                                                       oREF=filter_opts['oREF'], 
                                                                       laplacian_rename=filter_opts['laplacian_rename'], 
                                                                       renames=filter_opts['renames'],
                                                                       montage=filter_opts['montage']
                                                                       )
                                data.axis['chan'][0] = asarray([x for x in chanset])
                                selectchans = ch
                                dat = data[0]
                            else:
                                dat = data()[0][0]
                            
                            # If any errors occured during filtering, break
                            if filtflag > 0:
                                break
                            
                            # 7.g. Fix polarity of recording
                            if inversion:
                                dat = dat*-1 
           
                            # 7.h. Obtain phase signal
                            pha = squeeze(pac.filter(data.s_freq, dat, ftype='phase'))
                            
                            if len(pha.shape)>2:
                                pha = squeeze(pha)
                            
                            # 7.i. obtain amplitude signal
                            amp = squeeze(pac.filter(data.s_freq, dat, ftype='amplitude'))
                            if len(amp.shape)>2:
                                amp = squeeze(amp)
                            
                            # 7.j. extract signal (minus buffer)
                            nbuff = int(event_opts['buffer'] * data.s_freq)
                            minlen = data.s_freq * min_dur
                            if len(pha) >= 2 * nbuff + minlen:
                                pha = pha[nbuff:-nbuff]
                                amp = amp[nbuff:-nbuff]                               
                                
                            # Apply phase correction for hilbert transform
                            #pha = roll(pha, int(pi/2*s_freq), axis=-1)
        
                            # 7.k. put data in blocks (for surrogate testing)
                            longpha[s//50, s%50] = pha
                            longamp[s//50, s%50] = amp
                            
                            # 7.l. put data in long format (for preferred phase)
                            ampbin[s, :] = mean_amp(pha, amp, nbins=nbins)
        
                        # 8.a. if number of events not divisible by block length,
                        #    pad incomplete final block with randomly resampled events
                        sys.stdout.write('\r')
                        sys.stdout.flush()
                        
                        rem = len(seg) % 50
                        if rem > 0:
                            pads = 50 - rem
                            for pad in range(pads):
                                ran = random.randint(0,rem)
                                longpha[-1,rem+pad] = longpha[-1,ran]
                                longamp[-1,rem+pad] = longamp[-1,ran]
                        
                        # b. Calculate Coupling Strength
                        mi = zeros((longamp.shape[0],1))
                        mi_pv = zeros((longamp.shape[0],1))
                        for row in range(longamp.shape[0]): 
                            amp = zeros((1))   
                            pha = zeros((1)) 
                            for col in range(longamp.shape[1]):
                                if row == 62 and col == 25:
                                    print(f'Col: {col}')
                                pha = concatenate((pha,longpha[row,col]))
                                amp = concatenate((amp,longamp[row,col]))
                            pha = reshape(pha,(1,1,len(pha)))
                            amp = reshape(amp,(1,1,len(amp)))
                            mi[row] = pac.fit(pha, amp, n_perm=400,random_state=5,
                                          verbose=False)[0][0]
                            mi_pv[row] = pac.infer_pvalues(p=0.95, mcp='fdr')[0][0]
    
                        ## c. Calculate preferred phase
                        ampbin = ampbin / ampbin.sum(-1, keepdims=True) # normalise amplitude
                        ampbin = ampbin.squeeze()
                        ampbin = ampbin[~isnan(ampbin[:,0]),:] # remove nan trials
                        ab = ampbin
                        
                        # d. Create bins for preferred phase
                        vecbin = zeros(nbins)
                        width = 2 * pi / nbins
                        for n in range(nbins):
                            vecbin[n] = n * width + width / 2  
                        
                        # e. Calculate mean direction (theta) & mean vector length (rad)
                        ab_pk = argmax(ab,axis=1)
                        theta = circ_mean(vecbin,histogram(ab_pk,bins=nbins, 
                                                            range=(0,nbins))[0])
                        theta_deg = degrees(theta)
                        if theta_deg < 0:
                            theta_deg += 360
                        rad = circ_r(vecbin, histogram(ab_pk,bins=nbins)[0], d=width)
                        
                        # f. Take mean across all segments/events
                        ma = nanmean(ab, axis=0)
                        
                        # g. Correlation between mean amplitudes and phase-giving sine wave
                        sine = sin(linspace(-pi, pi, nbins))
                        sine = interp(sine, (sine.min(), sine.max()), (ma.min(), ma.max()))
                        rho, pv1 = circ_corrcc(ma, sine)
    
                        # h. Rayleigh test for non-uniformity of circular data
                        ppha = vecbin[ab.argmax(axis=-1)]
                        z, pv2 = circ_rayleigh(ppha)
                        pv2 = round(pv2,5)
                        
                        # 9.a Export and save data
                        if adap_bands_phase == 'Fixed':
                            phadap = '-fixed'
                        else:
                            phadap = '-adap'
                        if adap_bands_amplitude == 'Fixed':
                            ampadap = '-fixed'
                        else:
                            ampadap = '-adap'  
                        phaname1 = round(f_pha[0],2)
                        phaname2 = round(f_pha[1],2)
                        ampname1 = round(f_amp[0],2)
                        ampname2 = round(f_amp[1],2)
                        freqs = f'pha-{phaname1}-{phaname2}Hz{phadap}_amp-{ampname1}-{ampname2}Hz{ampadap}'
                        if model == 'whole_night':
                            stagename = '-'.join(self.stage)
                            outputfile = '{}/{}_{}_{}_{}_{}_{}_pac_parameters.csv'.format(
                                            outpath,sub,ses,fnamechan,stagename,'-'.join(evt_type),freqs)
                        elif model == 'stage*cycle':    
                            outputfile = '{}/{}_{}_{}_{}_cycle{}_{}_{}_pac_parameters.csv'.format(
                                          outpath,sub,ses,fnamechan,self.stage[nsg],cycle_idx[nsg],'-'.join(evt_type),freqs)
                        elif model == 'per_stage':
                            outputfile = '{}/{}_{}_{}_{}_{}_{}_pac_parameters.csv'.format(
                                          outpath,sub,ses,fnamechan,self.stage[nsg],'-'.join(evt_type),freqs)
                        elif model == 'per_cycle':
                            stagename = '-'.join(self.stage)
                            outputfile = '{}/{}_{}_{}_{}_cycle{}_{}_{}_pac_parameters.csv'.format(
                                          outpath,sub,ses,fnamechan,stagename,cycle_idx[nsg],'-'.join(evt_type),freqs)
    
                        # b. Save cfc metrics to dataframe
                        d = DataFrame([mean(pac.pac), mean(mi), median(mi_pv), theta, 
                                        theta_deg, rad, rho, pv1, z, pv2])
                        d = d.transpose()
                        d.columns = ['mi_raw','mi_norm','pval1','pp_radians','ppdegrees','mvl',
                                      'rho','pval2','rayl','pval3']
                        d.to_csv(path_or_buf=outputfile, sep=',')
                        
                        # c. Save binned amplitudes to pickle file
                        outputfile = outputfile.split('_pac_parameters.csv')[0] + '_mean_amps'
                        save(outputfile, ab)
       
        ### 10. Check completion status and print
        if flag == 0:
            logger.debug('Phase-amplitude coupling finished without ERROR.')  
        else:
            logger.warning('Phase-amplitude coupling finished with WARNINGS. See log for details.')
        
        #self.tracking = tracking   ## TO UPDATE - FIX TRACKING
        
        return                             



                        

def generate_adap_bands(peaks,width,chan):
    
    '''
    Generates adapted bands of 2 x width from a file containing spectral peaks,
    for the specified channels
    '''
    peaks1 = read_csv(peaks, delimiter=',',index_col=0)
    peaks2 = DataFrame(nan, index=peaks1.index, columns=peaks1.columns)

    
    for c,ch in enumerate(chan):
        peaks2[ch] =  [(x - 2.0, x + 2.0) for x in peaks1[ch]] 
         

    return peaks2    



def watson_williams(in_dir, out_dir, band_pairs, chan, cycle_idx, stage, nbins,
                    test = 'within', comps = [('all','V1'), ('all','V2')]):
    
    '''
    This script conducts a Watson-Williams test between two time-points (within)
    or between 2 groups (between)
            
    '''  
    
    if len(comps)>2:
        print('')
        print('Please only provide 2 comparisons at a time in comps.')
        
    else:
        
        # Setup output directory
        if path.exists(out_dir):
                print(out_dir + " already exists")
        else:
            mkdir(out_dir)
            
        
        # Check if band_pairs is a list
        if isinstance(band_pairs,str):
            band_pairs = [band_pairs]
            
        # Create output file
        dset = zeros((len(chan),len(band_pairs)*2))

        # Set vecbin
        width = 2 * pi / nbins
        vecbin = zeros(nbins)
        for i in range(nbins):
            vecbin[i] = i * width + width / 2
        
        # Loop through channels
        for k, ch in enumerate(chan):
            for b,bp in enumerate(band_pairs):
                
                print('')
                print(f'CHANNEL: {ch}')
                print(f'BAND PAIR: {bp}')
                print('')
                
                # Create output filename    
                stagename = ''.join(stage)
                partstr = ['_'.join(x) for x in comps]
                comparisons = [partstr[0], partstr[1]]
                bands = '_'.join(band_pairs)
                outname = '_vs_'.join([x for x in comparisons])
                filename = f'{stagename}_{bands}_{outname}'
                
                data_m = []
                # Loop through comparisons
                for c,(part,visit) in enumerate(comps):
                
                    # Set list of input participants & visits
                    if isinstance(part, list):
                        None
                    elif part == 'all':
                            part = listdir(in_dir)
                            part = [ p for p in part if not '.' in p]
                    else:
                        print('')
                        print("ERROR: comps must either contain a list of subject ids or = 'all' ")
                        print('')
                    part.sort()
                    for i, p in enumerate(part):
                        if visit == 'all':
                            visit = listdir(in_dir + '/' + p)
                            visit = [x for x in visit if not'.' in x]  
                    if isinstance(visit,str):
                        visit = [visit]
                    visit.sort()    
                    # Define output object   
                    datab = zeros((len(part),len(visit),nbins))
                    
                    # Loop through participants & visits
                    for i, p in enumerate(part):
                        for j, vis in enumerate(visit): 
                            if not path.exists(in_dir + '/' + p + '/' + vis + '/'):
                                print(f'WARNING: input folder missing for Subject {p}, visit {vis}, skipping..')
                                continue
                            else:
                                p_files = [s for s in listdir(in_dir + '/' + p + '/' + vis) if 
                                           bp in s if '.p' in s]
                                p_files = [s for s in p_files if stagename in s] 
                                p_files = [s for s in p_files if ch in s]
                                
                                # Open files containing mean amplitudes (if existing)
                                if len(p_files) == 0:
                                    print(f'WARNING: mean amplitudes file does not exist for Subject {p}, visit {vis}, stage {stagename}, channel {ch} - check this. Skipping..')
                                elif len(p_files) >1:
                                    print(f'WARNING: multiple mean amplitudes files exist for Subject {p}, visit {vis}, stage {stagename}, channel {ch} - check this. Skipping..')
                                else:
                                    print(f'Extracting... Subject {p}, visit {vis}')
                                    ab_file = in_dir + '/' + p + '/' + vis + '/' + p_files[0]
                                    with open(ab_file, 'rb') as f:
                                        ab = load(f)
                                    
                                    # Calculate mean amplitudes across night per subject
                                    ab = nanmean(ab, axis=0)
    
                                    # Caculate z-score for binned data
                                    datab[i,j,:] = zscore(ab)
                                 
                                    
                    # Remove nans from output and take average           
                    databz = array([[[x if not isnan(x) else 0 for x in dim1] 
                                     for dim1 in dim2] for dim2 in datab])
                    data = mean(databz,axis=1)
                    data_m.append(array([circ_mean(vecbin, data[x, :] * 1000) for x in 
                                  range(data.shape[0])]))
                    
                # Create array of data
    
                if test == 'within':
                    if len(data_m[0]) == len(data_m[1]):
                        # Run permutation testing
                        print('')
                        print("Running 10,000 permutations... ")
                        F = zeros((10000))
                        P = zeros((10000))
                        warnings = True
                        for pm in range(0,10000):
                            perm = random.choice(a=[False, True], size=(len(data_m[0])))
                            da = deepcopy(data_m[0])
                            db = deepcopy(data_m[1])
                            if pm>0:
                                da[perm] = data_m[1][perm]
                                db[perm] = data_m[0][perm]
                                warnings = False
                                
                            F[pm], P[pm] = circ_wwtest(da, db, ones(da.shape), 
                                                       ones(db.shape), warnings)
                        dset[k,b*2] = F[0]
                        dset[k,(b*2)+1] = sum(F>F[0])/len(F)
                    else:
                        print("For within-subjects comparisons, the number of subjects in each condition need to match... ")
                elif test == 'between':
                    da = deepcopy(data_m[0])
                    db = deepcopy(data_m[1])
                    F, P = circ_wwtest(da, db, ones(da.shape), ones(db.shape))
                    dset[k,b*2] = F
                    dset[k,(b*2)+1] = P
                else:
                    print("WARNING: test must either be 'between' or 'within' ... ")
            
            
        # Save output to file
        columns = [x+'_'+y for x in band_pairs for y in ['F','p']]
        df = DataFrame(dset, index=chan, columns=columns)
        df.to_csv(r"{out_dir}/watson_williams_{filename}.csv".format(out_dir=out_dir,
                                                           filename=filename))
            
            
            
        print('')
        print("Completed... ")

            

            




# -*- coding: utf-8 -*-
"""
Created on Thu Jul 29 10:29:11 2021

@author: Nathan Cross
"""
from copy import deepcopy
from csv import reader
from datetime import datetime
from fooof import FOOOF
from fooof.analysis import get_band_peak_fm
from glob import glob
from itertools import product
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from numpy import (append, arange, asarray, ceil, concatenate, empty, floor, 
                   mean, nan, ones, pi, reshape, sqrt, stack, sum, where, zeros)
from openpyxl import Workbook
from os import listdir, mkdir, path, walk
from pandas import DataFrame, ExcelWriter, read_csv
from scipy.fftpack import next_fast_len
from wonambi import ChanFreq, Dataset
from wonambi.attr import Annotations
from wonambi.trans import (fetch, frequency, get_descriptives, export_freq, 
                           export_freq_band)
from ..utils.misc import (bandpass_mne, laplacian_mne, notch_mne, notch_mne2, 
                          check_data_length)
from ..utils.logs import create_logger
from ..utils.load import infer_ref, load_channels, load_sessions, rename_channels


def default_general_opts():
    general_opts = {'freq_full':True, 
                    'freq_band':True, 
                    'freq_plot':False,
                    'max_freq_plot':35,
                    'suffix':None,
                    'chan_grp_name':'eeg'}
    return general_opts

def default_frequency_opts():
    headers = ['SO', 
               'Delta', 
               'Theta', 
               'Alpha', 
               'Sigma', 
               'Low Beta', 
               'High Beta']
    bands = [(0.25, 1.25),  #SO
             (0.5, 4),      #delta
             (4, 7.75),     #theta
             (8, 11),       #alpha
             (11.25, 16),   #sigma
             (16.25, 19),   #low beta
             (19.25, 35)]   #high beta
    frequency_opts = {'bands': bands,
                      'headers': headers,
                      'frequency': (11, 16),
                      'output':'spectraldensity', 
                      'scaling':'power', 
                      'sides':'one', 
                      'taper':'hann', 
                      'detrend':'linear', 
                      'n_fft': None, 
                      'fast_fft': True, 
                      'duration': 4,
                      'overlap': 0.5,
                      'step': None,
                      'centend':'mean',
                      'log_trans': False,
                      'halfbandwidth': 3,
                      'NW': None}
    return frequency_opts

def default_filter_opts():
    filter_opts = {'laplacian': False, 
                   'oREF': None, 
                   'lapchan': None, 
                   'laplacian_rename': False, 
                   'renames': None, 
                   'montage': 'standard_alphabetic',
                   'notch': True, 
                   'notch_freq': 50, 
                   'notch_harmonics': False,
                   'bandpass': True,
                   'highpass': 0.25,
                   'lowpass': 40,
                   'dcomplex': 'hilbert',
                   'filtcycle': (3, 6),
                   'width': 7}
    return filter_opts

def default_epoch_opts():
    epoch_opts = {'epoch': None,
                  'reject_epoch': True,
                  'reject_artf': True,
                  'min_dur':0,
                  'epoch_dur': 30,
                  'epoch_overlap': 0.5,
                  'epoch_step': None,
                  'concat_signal':True}
    
    return epoch_opts

def default_event_opts():
    event_opts = {'evt_type': None,
                  'event_chan': None,
                  'buffer': 2,
                  'concat_events':True}
    return event_opts

def default_fooof_opts():
    fooof_opts = {'psd_dur': 2, 
                  'peak_width_limits': [1, 12], 
                  'max_n_peaks': 8,
                  'min_peak_amplitude': 0.0,
                  'peak_threshold': 2.0,
                  'freq_range': [0.5, 35],
                  'select_highest': True,
                  'thresh_param': 'PW',
                  'bands_fooof': None, 
                  'thresh_select': None}
    return fooof_opts

def default_norm_opts():
    norm_opts = {'norm_cat': (1, 1, 1, 0),
                 'norm_evt_type': ['norm'], 
                 'norm_stage': None, 
                 'norm_epoch': None}
    return norm_opts


class Spectrum:
    
    """Design power spectral analyses on segmented data.

    Parameters
    ----------
    method : str
        one of the predefined methods
    frequency : tuple of float
        low and high frequency of frequency band
    duration : tuple of float
        min and max duration of spindles
    merge : bool
        if True, then after events are detected on every channel, events on 
        different channels that are separated by less than min_interval will be 
        merged into a single event, with 'chan' = the chan of the earlier-onset 
        event.
        
    Functions
    ----------
    fooof_it : 
        calculation of spectral parameters.
    powerspec_it : 
        Call to calculate power spectral analysis for each of:
            <sub>, <ses>, <chan>, <segment>
    powerspec_summary_full : 
        for power spectrum statistics. Returns value for whole spectrum, in 
        N/fs bins.
    powerspec_summary_bands : 
        for power spectrum statistics. Returns value for whole spectrum, 
        for pre-defined bands of interest.
        
    Notes
    -----
    See individual functions for other attribute descriptions.
    """ 
    
    def __init__(self, rec_dir, xml_dir, out_dir, chan, ref_chan, 
                 grp_name, stage, cat, rater = None, cycle_idx = None,
                 subs = 'all', sessions = 'all', tracking = None):
        
        self.rec_dir = rec_dir
        self.xml_dir = xml_dir
        self.out_dir = out_dir
        
        self.subs = subs
        self.sessions = sessions
        
        self.chan = chan
        self.ref_chan = ref_chan
        self.grp_name = grp_name
        self.rater = rater
        
        self.stage = stage
        self.cycle_idx = cycle_idx
        self.cat = cat
        
        if tracking == None:
            tracking = {}
        self.tracking = tracking
            

    
    def fooof_it(self, general_opts, frequency_opts, filter_opts, epoch_opts, 
                       event_opts, fooof_opts, filetype = '.edf', 
                       logger = create_logger('spectral peaks')):
        

        '''
        FOOOF is a fast, efficient, and physiologically-informed tool to 
        parameterize neural power spectra.
        
        Outputs
        -------
        ap_params: 
                  Parameters for the aperiodic component, reflecting 1/f like 
                  characteristics, including: 
                                    - offset
                                    - slope
        pk_params:
                  Parameters for periodic components (putative oscillations), 
                  as peaks rising above the aperiodic component, including:
                                    - peak frequency (in Hz)
                                    - peak bandwidth
                                    - peak amplitude (in µV^2)
        
        Notes
        -----
        See: https://fooof-tools.github.io/fooof/index.html for more info.
        
        '''
        
        ### 0.a. Set up logging
        logger.info('')
        tracking = self.tracking
        flag = 0
        
        ### 0.b. Set up organisation of export
        suffix = general_opts['suffix']
        freq_bw = frequency_opts['frequency']
        if self.cat[0] + self.cat[1] == 2:
            model = 'whole_night'
            logger.debug(f'Parameterizing power spectrum in range {freq_bw[0]}-{freq_bw[1]} Hz for the whole night.')
        elif self.cat[0] + self.cat[1] == 0:
            model = 'stage*cycle'
            logger.debug(f'Parameterizing power spectrum in range {freq_bw[0]}-{freq_bw[1]} Hz per stage and cycle separately.')
        elif self.cat[0] == 0:
            model = 'per_cycle'
            logger.debug(f'Parameterizing power spectrum in range {freq_bw[0]}-{freq_bw[1]} Hz per cycle separately.')
        elif self.cat[1] == 0:
            model = 'per_stage'  
            logger.debug(f'Parameterizing power spectrum in range {freq_bw[0]}-{freq_bw[1]} Hz per stage separately.')
        if 'cycle' in model and self.cycle_idx == None:
            logger.info('')
            logger.critical("To run cycles separately (i.e. cat[0] = 0), cycle_idx cannot be 'None'")
            return
        cat = tuple((self.cat[0],self.cat[1],1,1)) # force concatenation of discontinuous & events
        logger.info('')
        logger.debug(r"""
                     
                                  |
                                  |
                                  |  
                                  |  .
                                  |   `~.
                        (µV2)     |      `~.       FOOOF !
                                  |         `^.
                                  |            `~.
                                  |               `•._
                                  |                   `~¬.…_
                                  |_____________________________
                                              (Hz)
                                  
                                                    """,)
        
        # 1.a. Define model parameters
        if fooof_opts is None:
            logger.critical('Options not set for FOOOF')
            return
        else:
            fm = FOOOF(fooof_opts['peak_width_limits'], fooof_opts['max_n_peaks'], 
                       fooof_opts['min_peak_amplitude'], fooof_opts['peak_threshold'])
            def gaussian_integral(a, c):
                """ Returns definite integral of a gaussian function with height a and
                standard deviation c."""
                return sqrt(2) * a * abs(c) * sqrt(pi)
        
        # 1.b. Prepare Bands
        if fooof_opts['bands_fooof'] is not None:
            bands = fooof_opts['bands_fooof']
        else:
            stp = min(fooof_opts['peak_width_limits'])
            low = int(floor(fooof_opts['freq_range'][0]))
            hi = int(ceil(fooof_opts['freq_range'][1]))
            bands = [(x,x + stp) for x in range(low,hi)]
        
        # 2.a. Check for output folder, if doesn't exist, create
        if path.exists(self.out_dir):
                logger.debug(f"Output directory: {self.out_dir} exists")
        else:
            mkdir(self.out_dir)
        
        # 2.b. Get subjects
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = next(walk(self.xml_dir))[1]
        else:
            logger.info('')
            logger.error("'subs' must either be an array of Participant IDs or = 'all' ")
            return
        subs.sort()
        
        # 3.a. Begin loop through participants
        for p, sub in enumerate(subs):
            
            # b. Begin loop through sessions
            flag, sessions = load_sessions(sub, self.sessions, self.rec_dir, flag, 
                                     logger, verbose=2)

            for v, ses in enumerate(sessions):
                logger.info('')
                logger.debug(f'Commencing {sub}, {ses}')
                
                ## Define files
                rdir = f'{self.rec_dir}/{sub}/{ses}/eeg/'
                xdir = f'{self.xml_dir}/{sub}/{ses}/'
                
                try:
                    edf_file = [x for x in listdir(rdir) if x.endswith(filetype)]
                    dset = Dataset(rdir + edf_file[0])
                except:
                    logger.warning(f' No input {filetype} file in {rdir}. Skipping...')
                    flag+=1
                    break
                
                try:
                    xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]
                except:
                    logger.warning(f"No input annotations file in {xdir} or path doesn't exist. Skipping...")
                    flag+=1
                    break
                
                ### Define output path
                if not path.exists(self.out_dir):
                    mkdir(self.out_dir)
                if not path.exists(f'{self.out_dir}/{sub}'):
                    mkdir(f'{self.out_dir}/{sub}')
                if not path.exists(f'{self.out_dir}/{sub}/{ses}'):
                    mkdir(f'{self.out_dir}/{sub}/{ses}')
                outpath = f'{self.out_dir}/{sub}/{ses}'
                
                ## Now import data
                dset = Dataset(rdir + edf_file[0])
                annot = Annotations(xdir + xml_file[0], rater_name=self.rater)
                 
                ### get cycles
                if self.cycle_idx is not None:
                    all_cycles = annot.get_cycles()
                    cycle = [all_cycles[i - 1] for i in self.cycle_idx if i <= len(all_cycles)]
                else:
                    cycle = None
                    
                ### if event channel only, specify event channels
                # 4.d. Channel setup
                flag, chanset = load_channels(sub, ses, self.chan, 
                                              self.ref_chan, flag, logger)
                if not chanset:
                    flag+=1
                    break
                newchans = rename_channels(sub, ses, self.chan, logger)

                # get segments
                for c, ch in enumerate(chanset):
                    logger.debug(f"Reading data for {ch}:{'/'.join(chanset[ch])}")
                    segments = fetch(dset, annot, cat = cat, 
                                     evt_type = event_opts['evt_type'], 
                                     stage = self.stage, cycle=cycle,  
                                     epoch = epoch_opts['epoch'], 
                                     epoch_dur = epoch_opts['epoch_dur'], 
                                     epoch_overlap = epoch_opts['epoch_overlap'], 
                                     epoch_step = epoch_opts['epoch_step'], 
                                     reject_epoch = epoch_opts['reject_epoch'], 
                                     reject_artf = epoch_opts['reject_artf'],
                                     min_dur = epoch_opts['min_dur'])
                    
                    # 5.b Rename channel for output file (if required)
                    if newchans:
                        fnamechan = newchans[ch]
                    else:
                        fnamechan = ch
                    
                    # Read data
                    segments.read_data(chan = ch, ref_chan = chanset[ch])
                    if len(segments)==0:
                        logger.warning(f"No valid data found for {ch}:{'/'.join(chanset[ch])}.")
                        flag+=1
                        continue
                    
                    for sg, seg in enumerate(segments):
                        logger.debug(f"Analysing segment {sg+1} of {len(segments)} ({seg['stage']})")
                        out = dict(seg)
                        data = seg['data']
                        timeline = data.axis['time'][0]
                        out['start'] = timeline[0]
                        out['end'] = timeline[-1]
                        out['duration'] = len(timeline) / data.s_freq
                        
                        if frequency_opts['fast_fft']:
                            n_fft = frequency_opts['n_fft']
                        else:
                            n_fft = None
                    
                        Fooofxx = frequency(data, output=frequency_opts['output'], 
                                        scaling=frequency_opts['scaling'], 
                                        sides=frequency_opts['sides'], 
                                        taper=frequency_opts['taper'],
                                        halfbandwidth=frequency_opts['halfbandwidth'], 
                                        NW=frequency_opts['NW'],
                                        duration=frequency_opts['duration'], 
                                        overlap=frequency_opts['overlap'], 
                                        step=frequency_opts['step'],
                                        detrend=frequency_opts['detrend'], 
                                        n_fft=n_fft, 
                                        log_trans=False, 
                                        centend=frequency_opts['centend'])
                        
                        freqs = Fooofxx.axis['freq'][0]     
                        fooof_powers = zeros((len(bands)))
                        fooof_ap_params = zeros((len(bands), 2))
                        fooof_pk_params = ones((len(bands), 9)) * nan
                        fm.fit(freqs, Fooofxx.data[0][0], fooof_opts['freq_range'])
                        
                        for j, band in enumerate(bands):
                            fp = get_band_peak_fm(fm, band, fooof_opts['select_highest'],
                                                  threshold=fooof_opts['thresh_select'],
                                                  thresh_param=fooof_opts['thresh_param'])
                            if fp.ndim == 1:
                                fooof_powers[j] = gaussian_integral(fp[1], fp[2])
                            else:
                                pwr = asarray([gaussian_integral(fp[x, 1], fp[x, 2]) \
                                               for x in range(fp.shape[0])]).sum()
                                fooof_powers[j] = pwr
                                
                            # get fooof aperiodic parameters
                            fooof_ap_params[j, :] = fm.aperiodic_params_
                            
                            # get fooof peak parameters
                            fp = get_band_peak_fm(fm, band, False,
                                                  threshold=fooof_opts['thresh_select'],
                                                  thresh_param=fooof_opts['thresh_param'])
                            if fp.ndim == 1:
                                fooof_pk_params[j, :3] = fp
                            else:
                                n_peaks = min(fp.shape[0], 3)
                                fooof_pk_params[j, :n_peaks * 3] = fp[:n_peaks, 
                                                                      :].ravel()
                        out['fooof_powers'] = fooof_powers
                        out['fooof_ap_params'] = fooof_ap_params
                        out['fooof_pk_params'] = fooof_pk_params
                                
                        
    
                        seg_info = ['Start time', 'End time', 'Duration', 'Stitches', 
                                    'Stage', 'Cycle', 'Event type', 'Channel']
                        band_hdr = [f'{b1}-{b2} Hz' for b1, b2 in bands]
                        pk_params_hdr = ['peak1_CF', 'peak1_PW', 'peak1_BW', 
                                         'peak2_CF', 'peak2_PW', 'peak2_BW', 
                                         'peak3_CF', 'peak3_PW', 'peak3_BW', ]
                        ap_params_hdr = ['Offset', 'Exponent']
                        band_pk_params_hdr = ['_'.join((b, p)) for b in band_hdr 
                                              for p in pk_params_hdr]
                        band_ap_params_hdr = ['_'.join((b, p)) for b in band_hdr 
                                              for p in ap_params_hdr]
                        one_record = zeros((1, 
                                            (len(seg_info) + len(bands) + 
                                            len(band_pk_params_hdr) + 
                                            len(band_ap_params_hdr))),
                                            dtype='O')
                        one_record[0, :] = concatenate((asarray([
                                                                out['start'],
                                                                out['end'],
                                                                out['duration'], 
                                                                out['n_stitch'], # number of concatenated segments minus 1
                                                                out['stage'],
                                                                out['cycle'],
                                                                out['name'], # event type
                                                                ch,
                                                                ]),
                                                                out['fooof_powers'],
                                                                out['fooof_pk_params'].ravel(),
                                                                out['fooof_ap_params'].ravel(),
                                                                ))
                        
                    ### Output filename ###
                    if model == 'whole_night':
                        stagename = '-'.join(self.stage)
                        outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_specparams_{suffix}.csv'
                    elif model == 'stage*cycle':    
                        outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{self.stage[sg]}_cycle{self.cycle_idx[sg]}_specparams_{suffix}.csv'
                    elif model == 'per_stage':
                        outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{self.stage[sg]}_specparams_{suffix}.csv'
                    elif model == 'per_cycle':
                        stagename = '-'.join(self.stage)
                        outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_cycle{self.cycle_idx[sg]}_specparams_{suffix}.csv'

                    logger.debug(f'Saving {outputfile}')
                    df = DataFrame(data = one_record, 
                                   columns = (seg_info + band_hdr + band_pk_params_hdr
                                              + band_ap_params_hdr))
                    df.to_csv(outputfile)
                    
                    # Update tracking TO DO
                    #update = datetime.fromtimestamp(path.getmtime(outputfile)).strftime("%m-%d-%Y, %H:%M:%S")
        ### 3. Check completion status and print
        if flag == 0:
            logger.info('')
            logger.debug('Specparam finished without error.')  
        else:
            logger.info('')
            logger.warning(f'Specparam finished with {flag} WARNINGS. See log for details.')
        return 

    
    
    def powerspec_it(self, general_opts, frequency_opts, filter_opts, epoch_opts, 
                           event_opts, norm, norm_opts, filetype = '.edf',
                           logger = create_logger('Power Spectral Analysis')):
        
        ### 0.a. Set up logging
        tracking = self.tracking
        logger.info('')
        flag = 0
        
        ### 0.b. Set up organisation of export
        suffix = general_opts['suffix']
        freq_bw = frequency_opts['frequency']
        if self.cat[0] + self.cat[1] == 2:
            model = 'whole_night'
            logger.debug(f'Calculating power spectrum in range {freq_bw[0]}-{freq_bw[1]} '
                         'Hz for the whole night.')
        elif self.cat[0] + self.cat[1] == 0:
            model = 'stage*cycle'
            logger.debug(f'Calculating power spectrum in range {freq_bw[0]}-{freq_bw[1]} '
                         'Hz per stage and cycle separately.')
        elif self.cat[0] == 0:
            model = 'per_cycle'
            logger.debug(f'Calculating power spectrum in range {freq_bw[0]}-{freq_bw[1]} '
                         'Hz per cycle separately.')
        elif self.cat[1] == 0:
            model = 'per_stage'  
            logger.debug(f'Calculating power spectrum in range {freq_bw[0]}-{freq_bw[1]} '
                         'Hz per stage separately.')
        
        if self.cat[3] == 0:
            ev_model = True
            logger.debug(f"Calculating power spectrum in range {freq_bw[0]}-{freq_bw[1]} "
                         f"Hz per event: {event_opts['evt_type']} separately.")
        else:
            ev_model = False
        if 'cycle' in model and self.cycle_idx == None:
            logger.info('')
            logger.critical("To run cycles separately (i.e. cat[0] = 0), cycle_idx "
                            "cannot be 'None'")
            return
        if not ev_model and not epoch_opts['concat_signal']:
            logger.warning("If not running PSA per event, epoch_opts['concat_signal'] "
                           "must be set to True. Setting now."  )
            epoch_opts['concat_signal'] = True
        cat = self.cat
        #cat = tuple((self.cat[0],self.cat[1],1,1)) # force concatenation of discontinuous & events
        
        
        logger.info('')
        logger.debug(r"""
                     
                                  |
                                  |
                                  |  ,^\
                                  | /   \           Power Spectral
                                  |      `.            Analysis
                        (µV2)     |       `~.       
                                  |          `./\
                                  |              `~.
                                  |                 `•._
                                  |                     `~¬.…_._._
                                  |_______________________________
                                              (Hz)
                                  
                                                    """,)
        
        
        # 1.a. Check for output folder, if doesn't exist, create
        if path.exists(self.out_dir):
                logger.debug(f"Output directory: {self.out_dir} exists")
        else:
            mkdir(self.out_dir)
        
        # 2.b. Get subjects
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = next(walk(self.rec_dir))[1]
        else:
            logger.info('')
            logger.critical("'subs' must either be an array of Participant IDs or = 'all' ")
            return
        subs.sort()
        
        # 3.a. Begin loop through participants
        for p, sub in enumerate(subs):
            tracking[f'{sub}'] = {}
            
            # b. Begin loop through sessions
            sessions = self.sessions
            if sessions == 'all':
                sessions = listdir(self.rec_dir + '/' + sub)
                sessions = [x for x in sessions if not '.' in x] 

            for v, ses in enumerate(sessions):
                logger.info('')
                logger.debug(f'Commencing {sub}, {ses}')
                tracking[f'{sub}'][f'{ses}'] = {'powerspec':{}} 
                
                ## Define files
                rdir = f'{self.rec_dir}/{sub}/{ses}/eeg/'
                xdir = f'{self.xml_dir}/{sub}/{ses}/'
                
                try:
                    edf_file = [x for x in listdir(rdir) if x.endswith(filetype)]
                    dset = Dataset(rdir + edf_file[0])
                except:
                    logger.warning(f' No input {filetype} file in {rdir}. Skipping...')
                    break
                
                xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]            
                
                ### Define output path
                if not path.exists(self.out_dir):
                    mkdir(self.out_dir)
                if not path.exists(f'{self.out_dir}/{sub}'):
                    mkdir(f'{self.out_dir}/{sub}')
                if not path.exists(f'{self.out_dir}/{sub}/{ses}'):
                    mkdir(f'{self.out_dir}/{sub}/{ses}')
                outpath = f'{self.out_dir}/{sub}/{ses}'
                
                ## Now import data
                dset = Dataset(rdir + edf_file[0])
                annot = Annotations(xdir + xml_file[0], rater_name=self.rater)
                 
                ### get cycles
                if self.cycle_idx is not None:
                    all_cycles = annot.get_cycles()
                    cycle = [all_cycles[i - 1] for i in self.cycle_idx if i <= len(all_cycles)]
                else:
                    cycle = None
                    
                # 4.d. Channel setup          
                flag, chanset = load_channels(sub, ses, self.chan, 
                                              self.ref_chan, flag, logger)
                if not chanset:
                    logger.error('Problem loading channels from tracking sheet')
                    break
            
                newchans = rename_channels(sub, ses, self.chan, logger)
                if not filter_opts['oREF']:
                    filter_opts['oREF'] = infer_ref(sub, ses, self.chan, logger)    
                
                # Loop through channels
                for c, ch in enumerate(chanset):
                    
                    chan_full = [ch + ' (' + general_opts['chan_grp_name'] + ')']

                    # 5.b Rename channel for output file (if required)
                    if newchans:
                        fnamechan = newchans[ch]
                        #filter_opts['renames'] = {newchans[ch]:ch}
                        #filter_opts['laplacian_rename'] = True
                    else:
                        fnamechan = ch
                        #filter_opts['laplacian_rename'] = False
                        
                    if ch == '_REF':
                        filter_opts['laplacian_rename'] = False
                        if not filter_opts['oREF']:
                            try:
                                filter_opts['oREF'] = newchans[ch]
                            except:
                                logger.warning("Channel selected is '_REF' but "
                                               "no information has been given "
                                               "about what standard name applies, "
                                               "either via filter_opts['oREF'] "
                                               "or 'chanset_rename' in the tracking "
                                               "sheet (see user guide). Skipping "
                                               "channel...")
                                flag += 1
                                continue
                   
                    
                    # Normalisation (if requested)
                    if norm == 'baseline':
                        logger.debug('Extracting data for baseline normalisation...')
                        norm_seg = fetch(dset, annot, cat=norm_opts['norm_cat'], 
                                         evt_type=norm_opts['norm_evt_type'],
                                         stage=norm_opts['norm_stage'], 
                                         epoch=norm_opts['norm_epoch'])
                        if not norm_seg:
                            logger.warning('No valid baseline data found. Skipping...')
                            continue
                        if filter_opts['laplacian']:
                            try:
                                norm_seg.read_data(filter_opts['lapchan'], chanset[ch]) 
                                logger.debug("Applying Laplacian filtering to baseline data...")
                                laplace_flag = True
                            except:
                                logger.error(f"Channels listed in filter_opts['lapchan']: {filter_opts['lapchan']} are not found in recording.")
                                logger.warning("Laplacian filtering will NOT be run for BASELINE data. Check parameters under: filter_opts['lapchan']")
                                norm_seg.read_data(ch, chanset[ch])
                                laplace_flag = False
                                flag += 1
                        else:
                            norm_seg.read_data(ch, chanset[ch])            
                        all_nSxx = []
                        for seg in norm_seg:
                            normdata = seg['data']
                            if laplace_flag:
                                normdata.data[0] = laplacian_mne(normdata, 
                                                         filter_opts['oREF'], 
                                                         channel=chan_full, 
                                                         ref_chan=chanset[ch], 
                                                         laplacian_rename=filter_opts['laplacian_rename'], 
                                                         renames=filter_opts['renames'])
                            Sxx = frequency(normdata, output=frequency_opts['output'], 
                                            scaling=frequency_opts['scaling'],
                                            sides=frequency_opts['sides'], 
                                            taper=frequency_opts['taper'],
                                            halfbandwidth=frequency_opts['halfbandwidth'], 
                                            NW=frequency_opts['NW'],
                                            duration=frequency_opts['duration'], 
                                            overlap=frequency_opts['overlap'], 
                                            step=frequency_opts['step'],
                                            detrend=frequency_opts['detrend'], 
                                            n_fft=frequency_opts['n_fft'], 
                                            log_trans=frequency_opts['log_trans'], 
                                            centend=frequency_opts['centend'])
                            all_nSxx.append(Sxx)
                            nSxx = ChanFreq()
                            nSxx.s_freq = Sxx.s_freq
                            nSxx.axis['freq'] = Sxx.axis['freq']
                            nSxx.axis['chan'] = Sxx.axis['chan']
                            nSxx.data = empty(1, dtype='O')
                            nSxx.data[0] = empty((Sxx.number_of('chan')[0],
                                     Sxx.number_of('freq')[0]), dtype='f')
                            nSxx.data[0] = mean(
                                    stack([x()[0] for x in all_nSxx], axis=2), axis=2)
                    
                    
                    
                    ## Get segments of data
                    logger.debug(f"Reading data for {ch}:{'/'.join(chanset[ch])}")
                    
                    segments = fetch(dset, annot, cat = cat, 
                                     chan_full = [ch],
                                     evt_type = event_opts['evt_type'], 
                                     stage = self.stage, cycle=cycle,   
                                     epoch = epoch_opts['epoch'], 
                                     epoch_dur = epoch_opts['epoch_dur'], 
                                     epoch_overlap = epoch_opts['epoch_overlap'], 
                                     epoch_step = epoch_opts['epoch_step'], 
                                     reject_epoch = epoch_opts['reject_epoch'], 
                                     reject_artf = epoch_opts['reject_artf'],
                                     min_dur = epoch_opts['min_dur'],
                                     buffer = event_opts['buffer'])
                    if len(segments)==0:
                        logger.warning(f"No valid data found for {sub}, {ses}, {self.stage}, {cycle}.")
                        continue
                    
                    # Read data
                    if filter_opts['laplacian']:
                        try:
                            segments.read_data(filter_opts['lapchan'], chanset[ch]) 
                            logger.debug("Applying Laplacian filtering to data...")
                            laplace_flag = True
                        except:
                            logger.error(f"Channels listed in filter_opts['lapchan']: {filter_opts['lapchan']} are not found in recording.")
                            logger.warning("Laplacian filtering will NOT be run. Check parameters under: filter_opts['lapchan']")
                            segments.read_data(ch, chanset[ch])
                            laplace_flag = False
                            flag += 1
                    else:
                        segments.read_data(ch, chanset[ch])
                    
                    # Loop over segments and apply transforms
                    for sg, seg in enumerate(segments):
                        logger.debug(f"Analysing segment {sg+1} of {len(segments)} ({seg['stage']})")
                        out = dict(seg)
                        data = seg['data']
                        timeline = data.axis['time'][0]
                        out['start'] = timeline[0]
                        out['end'] = timeline[-1]
                        out['duration'] = len(timeline) / data.s_freq
                        
                        if not frequency_opts['fast_fft']:
                            n_fft = n_fft = 256*8
                        else:
                            n_fft = frequency_opts['n_fft']
                        
                        if filter_opts['laplacian']:
                            selectchans = filter_opts['lapchan']
                        else:
                            selectchans = ch
                        
                        # Notch filters
                        filtflag = 0
                        if filter_opts['notch']:
                            data.data[0], filtflag = notch_mne(data,  
                                                    channel=selectchans, 
                                                    freq=filter_opts['notch_freq'],
                                                    oREF=filter_opts['oREF'],
                                                    rename=filter_opts['laplacian_rename'],
                                                    renames=filter_opts['renames'],
                                                    montage=filter_opts['montage']
                                                    )
                            
                        if filter_opts['notch_harmonics']: 
                            data.data[0], filtflag = notch_mne2(data, 
                                                      channel=selectchans,
                                                      oREF=filter_opts['oREF'], 
                                                      rename=filter_opts['laplacian_rename'],
                                                      renames=filter_opts['renames'],
                                                      montage=filter_opts['montage']
                                                      )    
                        
                        # 7.e. Bandpass filters
                        if filter_opts['bandpass']:
                            data.data[0], filtflag = bandpass_mne(data, 
                                                        channel=selectchans, 
                                                        highpass=filter_opts['highpass'], 
                                                        lowpass=filter_opts['lowpass'], 
                                                        oREF=filter_opts['oREF'],
                                                        rename=filter_opts['laplacian_rename'],
                                                        renames=filter_opts['renames'],
                                                        montage=filter_opts['montage']
                                                        )
                        
                        # Laplacian transform
                        if filter_opts['laplacian'] and laplace_flag:
                            data.data[0], filtflag = laplacian_mne(data, 
                                                                   channel=selectchans, 
                                                                   ref_chan=chanset[ch], 
                                                                   oREF=filter_opts['oREF'], 
                                                                   laplacian_rename=filter_opts['laplacian_rename'], 
                                                                   renames=filter_opts['renames'],
                                                                   montage=filter_opts['montage']
                                                                   )
                            data.axis['chan'][0] = asarray([x for x in chanset])
                            selectchans = ch
                        
                        # If any errors occured during filtering, break
                        if filtflag > 0:
                            break
                        
                        ### Frequency transformation
                        try:
                            check_data_length(data, frequency_opts['duration'],
                                              logger)
                            Sxx = frequency(data, output=frequency_opts['output'], 
                                        scaling=frequency_opts['scaling'], 
                                        sides=frequency_opts['sides'], 
                                        taper=frequency_opts['taper'],
                                        halfbandwidth=frequency_opts['halfbandwidth'], 
                                        NW=frequency_opts['NW'],
                                        duration=frequency_opts['duration'], 
                                        overlap=frequency_opts['overlap'], 
                                        step=frequency_opts['step'],
                                        detrend=frequency_opts['detrend'], 
                                        n_fft=n_fft, 
                                        log_trans=frequency_opts['log_trans'], 
                                        centend=frequency_opts['centend'])
                        except Exception as error:
                            logger.error(error.args[0])
                            logger.warning(f"Skipping {seg['stage']} for channel {str(ch)} ... ")
                            flag+=1
                            continue
                            
                        
                        # Baseline normalisation
                        if norm:
                            logger.debug('Applying baseline normalisation')
                            for j, ch in enumerate(Sxx.axis['chan'][0]):
                                dat = Sxx.data[0][j,:]
                                sf = Sxx.axis['freq'][0]
                                f_res = sf[1] - sf[0] # frequency resolution
                                if norm == 'integral':
                                    norm_dat = sum(dat) * f_res # integral by midpoint rule
                                else:
                                    norm_dat = nSxx(chan=ch)[0]
                                Sxx.data[0][j,:] = dat / norm_dat
                        

                        out['data'] = Sxx

                        # Export and save data
                        ### Output filename ###
                        if not ev_model:
                            if model == 'whole_night':
                                stagename = '-'.join(self.stage)
                                outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}**{suffix}.csv'
                            elif model == 'stage*cycle':    
                                outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{self.stage[sg]}_cycle{self.cycle_idx[sg]}**{suffix}.csv'
                            elif model == 'per_stage':
                                outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{self.stage[sg]}**{suffix}.csv'
                            elif model == 'per_cycle':
                                stagename = '-'.join(self.stage)
                                outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_cycle{self.cycle_idx[sg]}**{suffix}.csv'
                            if general_opts['freq_full']:
                                save_freq([out], freq_bw, general_opts, outputfile, logger)
                            if general_opts['freq_band']:
                                save_bands([out], freq_bw, general_opts, frequency_opts, outputfile, logger)
                            if general_opts['freq_plot']:
                                save_plot([out], general_opts, outputfile, logger)
                        elif sg == 0:
                            out_full = [deepcopy(out)]
                        else:
                            out_full.append(out)
                    
                    # If analyses set to be per event        
                    if ev_model:
                        ev_out = []
                        outputfile = [] #f"{outputfile.split('*')[0]}_{event_opts['evt_type'][0]}_**{outputfile.split('*')[2]}"
                        if model == 'whole_night':
                            outputfile.append(f"{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_{event_opts['evt_type'][0]}**{suffix}.csv")
                            ev_out.append(out_full)
                        if model == 'stage*cycle':
                            for (stage,cycle) in product(self.stage,self.cycle_idx):
                                outputfile.append(f"{outpath}/{sub}_{ses}_{fnamechan}_{stage}_cycle{cycle}_{event_opts['evt_type'][0]}**{suffix}.csv")
                                out = [x for x in out_full if stage == x['stage']]
                                out = [x for x in out if cycle in x['cycle']]
                                ev_out.append(out)
                        if model == 'per_stage':
                            for stage in self.stage:
                                outputfile.append(f"{outpath}/{sub}_{ses}_{fnamechan}_{stage}_{event_opts['evt_type'][0]}**{suffix}.csv")
                                out = [x for x in out_full if stage == x['stage']]
                                ev_out.append(out)
                        if model == 'per_cycle':
                            for cycle in self.cycle_idx:
                                outputfile.append(f"{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_cycle{cycle}_{event_opts['evt_type'][0]}**{suffix}.csv") 
                                out = [x for x in out if cycle in x['cycle']]
                                ev_out.append(out)
                        for out, outfile in zip(ev_out, outputfile):
                            if general_opts['freq_full']:
                                save_freq(out, freq_bw, general_opts, outfile, logger)
                            if general_opts['freq_band']:
                                save_bands(out, freq_bw, general_opts, frequency_opts, outfile, logger)
                            if general_opts['freq_plot']:
                                save_plot(out, general_opts, outfile, logger)     
                        
        ### 3. Check completion status and print
        if flag == 0:
            logger.debug('Power spectral analyses finished without error.')  
        else:
            logger.warning(f'Power spectral analyses finished with {flag} WARNINGS. See log for details.')
        
        #self.tracking = tracking   ## TO UPDATE - FIX TRACKING
        
        return 
           
        
    def powerspec_summary(self, chan, general_opts, frequency_opts, filter_opts, 
                                epoch_opts, event_opts, logger):
        
        ### 0.a. Set up logging
        logger = create_logger('Power Spectral Summary')
        logger.info('')
        
        logger.info('')
        logger.debug(r""" Summarising Power Spectrum Data...
                     
                                     .^.
                                    / |\\_
                                   / | \ ¯;
                                   `|  \ /
                                    |  \ \
                                    |  \ \
                                    \__\_|
                                    (    )
                                    (0 ) )
                                   // ||\\
                                 /(( // ||
                               // \\))||\\
                              )) //|| ||))
                                (( )) |//
                                  //  ((
                                  
                    Spectral Quantification Unified In a Dataset
                    (S.Q.U.I.D)
                    
                                                    """,)
        
        ### 0.b. Set up output file params
        suffix = general_opts['suffix']
        frequency = frequency_opts['frequency']
        flag = 0
        if general_opts['freq_full']:
            variables_full = [str(x) for x in (arange(0.25,filter_opts['lowpass'],0.25))]
            idx_data_full = list(range(9,8+filter_opts['lowpass']*4))
        if general_opts['freq_band']:
            col_headers = frequency_opts['headers'] 
            variables_band = col_headers
            idx_data_band = list(range(9,9+len(col_headers)))
        
        ### 0.c. Set up organisation of export
        if self.cat[0] == 0 and self.cycle_idx == None:
            logger.info('')
            logger.critical("To run cycles separately (i.e. cat[0] = 0), cycle_idx cannot be 'None'")
            return
        if self.cat[0] + self.cat[1] == 2:
            model = 'whole_night'
            logger.debug(f'Summarising power spectrum in range {frequency[0]}-{frequency[1]} Hz for the whole night.')
        elif self.cat[0] + self.cat[1] == 0:
            model = 'stage*cycle'
            logger.debug(f'Calculating power spectrum in range {frequency[0]}-{frequency[1]} Hz per stage and cycle separately.')
        elif self.cat[0] == 0:
            model = 'per_cycle'
            logger.debug(f'Calculating power spectrum in range {frequency[0]}-{frequency[1]} Hz per cycle separately.')
        elif self.cat[1] == 0:
            model = 'per_stage'  
            logger.debug(f'Calculating power spectrum in range {frequency[0]}-{frequency[1]} Hz per stage separately.')
        
        ### 1. First check the directories
        # a. Check for output folder, if doesn't exist, create
        if not path.exists(self.out_dir):
            mkdir(self.out_dir)
        out_dir = f'{self.out_dir}/powerspectrum_{model}'
        if path.exists(out_dir):
            logger.debug(f"Output directory: {out_dir} exists")
        else:
            logger.debug(f"Creating output directory: {out_dir}")
            mkdir(out_dir)
        
        # b. Get subject IDs
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = next(walk(self.xml_dir))[1]
        else:
            logger.error("'subs' must either be an array of participant ids or = 'all' ")
            
        # c. Take a look through directories to get sessions
        subs.sort()
        sessions = {}
        for s, sub in enumerate(subs):
            if self.sessions == 'all':
                sessions[sub] = next(walk(f'{self.xml_dir}/{sub}'))[1]
            elif isinstance(self.sessions, list):
                sessions = self.sessions
                  
        if isinstance(sessions, dict):
            sessions = list(set([y for x in sessions.values() for y in x])) 
            
        sessions.sort() 
  
        # 4. Begin data extraction
        for c, ch in enumerate(chan):
            logger.debug(f'Creating a summary dataset for channel: {ch}')
            
            # a. Create column names (append chan and ses names)
            for v, ses in enumerate(sessions):
                if general_opts['freq_full']:
                    sesvar = []
                    for pair in product(variables_full, [ses]):
                        sesvar.append('_'.join(pair))
                    columns_f = []
                    for pair in product(sesvar, [ch]):
                        columns_f.append('_'.join(pair))
                if general_opts['freq_band']:
                    sesvar = []
                    for pair in product(variables_band, [ses]):
                        sesvar.append('_'.join(pair))
                    columns_b = []
                    for pair in product(sesvar, [ch]):
                        columns_b.append('_'.join(pair))
                        
                # b. Extract data based on cycle and stage setup
                if model == 'whole_night':
                    stagename = '-'.join(self.stage)
                    if general_opts['freq_full']:
                        st_columns = [x + f'_{stagename}' for x in columns_f]
                        freq_full = DataFrame(index=subs, columns=st_columns, dtype=float)
                    if general_opts['freq_full']:    
                        st_columns = [x + f'_{stagename}' for x in columns_b]
                        freq_band = DataFrame(index=subs, columns=st_columns, dtype=float)
                    for sub in subs: 
                        logger.debug(f'Extracting from {sub}, {ses}')
                        paramsfile = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{stagename}**{suffix}.csv'
                        if general_opts['freq_full']:
                            filename = f"{paramsfile.split('*')[0]}_freq_full_{paramsfile.split('*')[2]}"
                            if not path.exists(filename):
                                logger.warning(f"No (exported) spectrum data found for {sub}, {ses}. Has 'power_spectrum' by 'freq_full' been run for {model}? Skipping...")
                                flag += 1
                                continue
                            else:
                                try:
                                    df = read_csv(filename, skiprows=1)
                                    freq_full.loc[sub] = df.iloc[-1,idx_data_full].to_numpy()
                                except:
                                    extract_psa_error(logger)
                                    flag +=1
                                    continue
                        if general_opts['freq_band']:
                            filename = f"{paramsfile.split('*')[0]}_freq_band_{paramsfile.split('*')[2]}" 
                            if not path.exists(filename):
                                logger.warning(f"No (exported) spectrum data found for {sub}, {ses}. Has 'power_spectrum' by 'freq_band' been run for {model}? Skipping...")
                                flag += 1
                                continue
                            else:
                                try:
                                    df = read_csv(filename, skiprows=1)
                                    freq_band.loc[sub] = df.iloc[-1,idx_data_band].to_numpy()
                                except:
                                    extract_psa_error(logger)
                                    flag +=1
                                    continue
                    if general_opts['freq_full']:
                        freq_full.to_csv(f"{out_dir}/powerspectrum_{ses}_{ch}_{stagename}_freq_full_{suffix}.csv")
                    if general_opts['freq_band']:
                        freq_band.to_csv(f"{out_dir}/powerspectrum_{ses}_{ch}_{stagename}_freq_band_{suffix}.csv")
 
                elif model == 'stage*cycle':
                    for cyc in self.cycle_idx:
                        cycle = f'cycle{cyc}'
                        for st in self.stage:
                            if general_opts['freq_full']:
                                st_columns = [x + f'_{st}_{cycle}' for x in columns_f]
                                freq_full = DataFrame(index=subs, columns=st_columns, dtype=float)
                            if general_opts['freq_full']:    
                                st_columns = [x + f'_{st}_{cycle}' for x in columns_b]
                                freq_band = DataFrame(index=subs, columns=st_columns, dtype=float) 
                            for sub in subs: 
                                logger.debug(f'Extracting from {sub}, {ses}, stage {st}, {cycle}')
                                paramsfile = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{st}_cycle{cycle}**{suffix}.csv'
                                if general_opts['freq_full']:
                                    filename = f"{paramsfile.split('*')[0]}_freq_full_{paramsfile.split('*')[2]}"
                                    if not path.exists(filename):
                                        logger.warning(f"No (exported) spectrum data found for {sub}, {ses}. Has 'power_spectrum' by 'freq_full' been run for {model}? Skipping...")
                                        flag += 1
                                        continue
                                    else:
                                        try:
                                            df = read_csv(filename, skiprows=1)
                                            freq_full.loc[sub] = df.iloc[-1,idx_data_full].to_numpy()
                                        except:
                                            extract_psa_error(logger)
                                            flag +=1
                                            continue
                                if general_opts['freq_band']:
                                    filename = f"{paramsfile.split('*')[0]}_freq_band_{paramsfile.split('*')[2]}" 
                                    if not path.exists(filename):
                                        logger.warning(f"No (exported) spectrum data found for {sub}, {ses}. Has 'power_spectrum' by 'freq_band' been run for {model}? Skipping...")
                                        flag += 1
                                        continue
                                    else:
                                        try:
                                            df = read_csv(filename, skiprows=1)
                                            freq_band.loc[sub] = df.iloc[-1,idx_data_band].to_numpy()
                                        except:
                                            extract_psa_error(logger)
                                            flag +=1
                                            continue
                            if general_opts['freq_full']:
                                freq_full.to_csv(f"{out_dir}/pectrum_{ses}_{ch}_{st}_{cycle}_freq_full_{suffix}.csv")
                            if general_opts['freq_band']:
                                freq_band.to_csv(f"{out_dir}/spectrum_{ses}_{ch}_{st}_{cycle}_freq_band_{suffix}.csv")

                elif model == 'per_cycle':
                    for cyc in self.cycle_idx:
                        cycle = f'cycle{cyc}'
                        stagename = '-'.join(self.stage)
                        if general_opts['freq_full']:
                            st_columns = [x + f'_{stagename}_{cycle}' for x in columns_f]
                            freq_full = DataFrame(index=subs, columns=st_columns, dtype=float)
                        if general_opts['freq_full']:    
                            st_columns = [x + f'_{stagename}_{cycle}' for x in columns_b]
                            freq_band = DataFrame(index=subs, columns=st_columns, dtype=float) 
                        for sub in subs: 
                            logger.debug(f'Extracting from {sub}, {ses}, {cycle}')
                            paramsfile = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{stagename}_{cycle}**{suffix}.csv'
                            if general_opts['freq_full']:
                                filename = f"{paramsfile.split('*')[0]}_freq_full_{paramsfile.split('*')[2]}"
                                if not path.exists(filename):
                                    logger.warning(f"No (exported) spectrum data found for {sub}, {ses}. Has 'power_spectrum' by 'freq_full' been run for {model}? Skipping...")
                                    flag += 1
                                    continue
                                else:
                                    try:
                                        df = read_csv(filename, skiprows=1)
                                        freq_full.loc[sub] = df.iloc[-1,idx_data_full].to_numpy()
                                    except:
                                        extract_psa_error(logger)
                                        flag +=1
                                        continue
                            if general_opts['freq_band']:
                                filename = f"{paramsfile.split('*')[0]}_freq_band_{paramsfile.split('*')[2]}" 
                                if not path.exists(filename):
                                    logger.warning(f"No (exported) spectrum data found for {sub}, {ses}. Has 'power_spectrum' by 'freq_band' been run for {model}? Skipping...")
                                    flag += 1
                                    continue
                                else:
                                    try:
                                        df = read_csv(filename, skiprows=1)
                                        freq_band.loc[sub] = df.iloc[-1,idx_data_band].to_numpy()
                                    except:
                                        extract_psa_error(logger)
                                        flag +=1
                                        continue
                        if general_opts['freq_full']:
                            freq_full.to_csv(f"{out_dir}/spectrum_{ses}_{ch}_{stagename}_{cycle}_freq_full_{suffix}.csv")
                        if general_opts['freq_band']:
                            freq_band.to_csv(f"{out_dir}/spectrum_{ses}_{ch}_{stagename}_{cycle}_freq_band_{suffix}.csv")

                elif model == 'per_stage':
                    for st in self.stage:
                        if general_opts['freq_full']:
                            st_columns = [x + f'_{st}' for x in columns_f]
                            freq_full = DataFrame(index=subs, columns=st_columns, dtype=float)
                        if general_opts['freq_full']:    
                            st_columns = [x + f'_{st}' for x in columns_b]
                            freq_band = DataFrame(index=subs, columns=st_columns, dtype=float) 
                        for sub in subs: 
                            logger.debug(f'Extracting from {sub}, {ses}, stage {st}')
                            paramsfile = f'{self.xml_dir}/{sub}/{ses}/{sub}_{ses}_{ch}_{st}**{suffix}.csv'
                            if general_opts['freq_full']:
                                filename = f"{paramsfile.split('*')[0]}_freq_full_{paramsfile.split('*')[2]}"
                                if not path.exists(filename):
                                    logger.warning(f"No (exported) spectrum data found for {sub}, {ses}. Has 'power_spectrum' by 'freq_full' been run for {model}? Skipping...")
                                    flag += 1
                                    continue
                                else:
                                    try:
                                        df = read_csv(filename, skiprows=1)
                                        freq_full.loc[sub] = df.iloc[-1,idx_data_full].to_numpy()
                                    except:
                                        extract_psa_error(logger)
                                        flag +=1
                                        continue
                            if general_opts['freq_band']:
                                filename = f"{paramsfile.split('*')[0]}_freq_band_{paramsfile.split('*')[2]}" 
                                if not path.exists(filename):
                                    logger.warning(f"No (exported) spectrum data found for {sub}, {ses}. Has 'power_spectrum' by 'freq_band' been run for {model}? Skipping...")
                                    flag += 1
                                    continue
                                else:
                                    try:
                                        df = read_csv(filename, skiprows=1)
                                        freq_band.loc[sub] = df.iloc[-1,idx_data_band].to_numpy()
                                    except:
                                        extract_psa_error(logger)
                                        flag +=1
                                        continue
                        if general_opts['freq_full']:
                            freq_full.to_csv(f"{out_dir}/spectrum_{ses}_{ch}_{st}_freq_full_{suffix}.csv")
                        if general_opts['freq_band']:
                            freq_band.to_csv(f"{out_dir}/spectrum_{ses}_{ch}_{st}_freq_band_{suffix}.csv")
                            
         
        ### 3. Check completion status and print
        if flag == 0:
            logger.info('')
            logger.debug('Create powerspec dataset finished without error.')  
        else:
            logger.info('')
            logger.warning(f'Create powerspec dataset finished with {flag} WARNINGS. See log for details.')
        return 
    
                        
# MISC FUNCTIONS        
def extract_psa_error(logger):
    logger.critical("Data extraction error: Check that all 'params' are written correctly.")
    logger.info('                      Check documentation for how event parameters need to be written:')
    logger.info('                      https://seapipe.readthedocs.io/en/latest/index.html')            
        
        
def save_freq(out, freq_bw, general_opts, outputfile, logger):

    low = where(out[0]['data'].axis['freq'][0]==freq_bw[0])[0][0]
    hi = where(out[0]['data'].axis['freq'][0]==freq_bw[1])[0][0]
    
    #[x[low:hi] for x['data'].axis['freq'][0] in out]
    for i, x in enumerate(out):
        out[i]['data'].axis['freq'][0] = x['data'].axis['freq'][0][low:hi]
        out[i]['data'].data[0] = reshape(out[i]['data'].data[0][0][low:hi], 
                                         (1, len(out[i]['data'].data[0][0][low:hi])))
    as_matrix = asarray([x['data'].data[0][0] for x in out])
    # out['data'].axis['freq'][0] = out['data'].axis['freq'][0][low:hi]
    # out['data'].data[0] = asarray([out['data'].data[0][0][low:hi]])
    #as_matrix = asarray([y for y in out[0]['data'].data[0]])
    desc = get_descriptives(as_matrix)
    filename = f"{outputfile.split('*')[0]}_freq_full_{outputfile.split('*')[2]}"
    logger.debug(f'Writing to {filename}')  
    export_freq(out, filename, desc=desc)
        
def save_bands(out, freq_bw, general_opts, frequency_opts, outputfile, logger):
    filename = f"{outputfile.split('*')[0]}_freq_band_{outputfile.split('*')[2]}"
    logger.debug(f'Writing to {filename}') 
    try:
        export_freq_band(out, frequency_opts['bands'], filename)
    except:
        logger.error('Cannot export power in user-defined frequency bands. Check bands. For info on how to define frequency bands, refer to documentation:')
        logger.info('https://seapipe.readthedocs.io/en/latest/index.html')

def save_plot(out, general_opts, outputfile, logger):
    
    ''' TO COMPLETE '''
    
    # fig = Figure()
    # FigureCanvas(fig)
    # idx_plot = 1
    # data = asarray([x['data'].data[0][0] for x in out])
    # seg_chan = data[0].axis['chan'][0]
    # sf = out[0]['data'].axis['freq'][0]
    # if general_opts['max_freq_plot']:
    #     idx_max = asarray(
    #             [abs(x - general_opts['max_freq_plot']) for x in sf]).argmin()
    # for ch in seg_chan:
    #     Sxx = data(chan=ch)[0]
    #     ax = fig.add_subplot(len([out]), len(seg_chan), idx_plot)
    #     ax.semilogy(sf[1:idx_max], Sxx[1:idx_max])
    #     ax.set_xlabel('Frequency (Hz)')
    #     idx_plot += 1 
    # filename = f"{outputfile.split('*')[0]}_powerspectrum.svg"
    # logger.debug(f'Saving figure to {filename}')
    # fig.savefig(f'{filename}')  
        
        